# ****************************************************
# Name des Moduls: BBox
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Thorsten Reichelt, Thorsten.Reichelt@web.de
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 10.01.2003
# ****************************************************


""" Diese Bibliothek erzeugt verschiedene Arten von BoundingBoxes, d.h., kleinstmglichen
    Rechtecken, die ein GeoObject (also ein Polygon) vollstndig beinhalten. Eine BB ist die
    Grundlage vieler Berechnungen. Verschiedene BBs gibt es, da unterschiedliche Sichtweisen
    (intrinsisch, deiktisch, geoglobal) fr die Berechnung der Prpositionsflchen notwendig
    sind und jeweils andere BBs erzeugen. Fr eine ausfhrlichere Erklrung des zugrundeliegen-
    den Konzepts siehe die allgemeine Beschreibung des Projekts und der der Bufferberechnungs-
    algorithmen.

    Die Basisklasse ist BBox (abgeleitet von geometry.Polygon). Mit dieser Klasse kann unter
    anderem die sog. Default-BB berechnet werden, deren Ausrichtung sich nach der lngsten
    Linie des fraglichen Polygons richtet.

    Von BBox abgeleitet sind:
        BBintrinsic
        BBdeictic und
        BBgeo
    fr intrinsisch, deiktisch und geoglobal ausgerichtete BoundingBoxen. Zur Erluterung
    dieses Vokabulars siehe bitte ebenfalls die allgemeine Projektbeschreibung.

    Die eigentliche Berechnung der BBs bernimmt die Funktion getBoundingBox() aus Titus
    von der Malsburgs Modul geobuffer. Aufgabe von BBox ist es, das Polygon, fr das eine
    BB berechnet werden soll, vorher so im Koordinatensystem zu drehen, dass das richtige
    Ergebnis dabei herauskommt...
"""    

from geometry import *
from geobuffer import getBoundingBox
from btools import indexcycle, listrotate
from berror import GeometryError, ParameterError, SemanticError
import copy

class BBox(Polygon):

    def __init__(self, geoobj, direction= ""):
        """ Die Berechnungen fr die verschiedenen Arten einer BB hneln sich sehr und
            knnen fr alle Typen gemeinsam durchgefhrt werden. Es muss ein Polygon
            (geoobj) bergeben werden, fr das die BB berechnet werden soll. Wird kein
            Richtungsvektor angegeben (direction, vom Typ geometry.Line), nach dem die
            Bounding-Box ausgerichtet werden soll, wird die lngste Linie des Objekts
            verwendet.
        """

        if not direction:
            #Lngste Linie des Objekts bestimmen
            l= 0
            i= 0
            line= 0
            for x in geoobj.lines:
                if x.length() > l:
                    l= x.length()
                    line= i
                i+= 1

            #direction= lngste gefundene Linie
            direction= geoobj.lines[line]
                    
  
        #BB berechnen
        bb= self.calculateBox(geoobj, direction)
        
        #Konstruktor des Polygons aufrufen
        id= 0
        Polygon.__init__(self, id, bb.points)
        

    def calculateBox(self, geoobj, direction):
        """ Berechnet eine nach direction ausgerichtete BB fr geoobj """

        id= 0
            
        #Winkel zwischen Richtungsvektor und Y-Achse berechnen...

        #Line auf der y-Achse
        l2= Line(id, Point(id,0,0), Point(id,0,1))
            
        #... der Winkel ist dann:
        alpha= angleClockwise(direction, l2)

        #Objektkopie um alpha drehen (Standard-BoundingBox ist immer parallel zu den Koordinatenachsen)
        cObj= copy.deepcopy(geoobj)
        cObj.rotate(alpha)

        #BoundingBox zum gedrehten Objekt erzeugen und zurckdrehen
        box= getBoundingBox(cObj)
        box.rotate(360- alpha)

        return box
    

    def setfirstside(self, v):
        """ Legt die Linie als erste Seite der BoundingBox fest, die von v (eine Linie) zuerst(!)
            geschnitten wird. Das ist wichtig, da die Algorithmen einiger Prpositionen mit den
            Seiten der BB arbeiten. So entstehen die Nebenflchen zum Beispiel auf den Seiten 2
            und 4 einer BB. Wenn nun vektor auf die Vorderseite eines Objekts zeigt, kann man nach
            der Anwendung dieser Methode sagen, welches denn nun die Seiten 2 und 4 des
            Objekts sind.
        """

        vektor= copy.deepcopy(v)        
        #Da vektor in Wahrheit nur eine Linie ist, mssen wir ihn vielleicht ein bisschen strecken
        #Also erstmal den grten Abstand des Anfangspunkts zur BB bestimmen
        distanceBB= 0
        for x in self.points:
            d= DPointPoint(x, vektor.P1)
            if d > distanceBB:
                distanceBB= d

        #Nun unseren "Vektor" auf die richtige Lnge skalieren
        vektor.scaleToLength(distanceBB+1)
        
        #Jetzt rausfinden, welche BB-Seite er zuerst schneidet
        distance= distanceBB
        firstside= 0
        i= 0
        
        for l in self.lines:
            if intersects(l, vektor):
                #Wenn es einen Schnittpunkt gibt, dessen Entfernung zum ersten Punkt des Vektors bestimmen
                d= DPointLine(vektor.P1, l)                
                if d < distance:
                    distance= d
                    firstside= i
            i+= 1

        #Spezialfall deiktische BB:
        #wenn sich der Viewer innerhalb der BB befindet, muss die BB-Seite zur ersten Seite
        #gemacht werden, die der geschnittenen Seite gegenberliegt
        if isinstance(self, BBdeictic) and contains(v.P1, self):
            firstside = indexcycle(firstside +2, 4)
    
        if distance > 0:
            #Die ermittelte Seite zur ersten Seite der BB machen
            self.switchside(firstside)    
        else:
            #Fehler! Kein Schnittpunkt gefunden
            raise GeometryError("BBox.setfirstside", 1)
        

    def switchside(self, firstside):
        """ Dreht die BB so, da die mit firstside indizierte Seite die erste ist """
        nlines= []
        npoints= []
        line= firstside
        for x in range(len(self.lines)):
            nlines.append(self.lines[line])
            npoints.append(self.lines[line].P2)

            line+= 1
            if line == len(self.lines):
                line= 0

        self.lines= nlines
        self.points= npoints


    def getcenter(self):
        """ Gibt den Mittelpunkt der BB zurck """

        xdiff = (self.points[0].x - self.points[2].x) / 2
        ydiff = (self.points[0].y - self.points[2].y) / 2
        id= 0
        
        return Point(id, self.points[0].x - xdiff, self.points[0].y - ydiff)

    
class BBintrinsic(BBox):

    def __init__(self, geoobj, direction):
        """ Erzeugt eine intrinsische BB. Es mssen ein Polygon (geoobj) und eine
            Richtungslinie (direction) angegeben werden. Damit die Berechnung
            vernnftige Ergebnisse liefert, muss sich zumindest der erste Punkt
            der Richtungslinie innerhalb von geoobj befinden. Die Richtungslinie wird
            als Vektor vom ersten zum zweiten Punkt interpretiert, der auf die Vorder-
            seite des Objekts zeigt.
        """

        #berprfen, ob in direction eine Linie bergeben wurde        
        if not isinstance(direction, Line):
            raise ParameterError("BBintrinsic.init", 1)
            
        #Konstruktor von BBox aufrufen
        BBox.__init__(self, geoobj, direction)
        
        #Erste Seite der BB = die Seite, die direction zuerst schneidet
        self.setfirstside(direction)


class BBdeictic(BBox):

    def __init__(self, geoobj, viewer):
        """ Erzeugt eine deiktische BB. Es mssen ein Polygon (geoobj) und ein
            Punkt (viewer) angegeben werden (Instanzen von geoobject.ViewerObject
            sind legal, da auch von geometry.Point abgeleitet). Die Ausrichtung der
            BB erfolgt nach einem Vektor, der von viewer zum Mittelpunkt des Polygons
            zeigt. Da der Mittelpunkt eines Polygons mitunter schwierig zu bestimmen
            ist, wird in Wahrheit der Mittelpunkt einer Default-BB verwendet ;-)
        """

        #Auf semantischen Fehler berprfen: Befindet sich der Viewer im fraglichen Objekt?
        if contains(viewer, geoobj):
            raise SemanticError("BBdeictic.init", 1)
               
        #und jetzt der Normalfall...
        else:
            #Den "Mittelpunkt" des Objekts bestimmen
            simpleBB= BBox(geoobj)
            center= simpleBB.getcenter()

            #Dahin wollen wir gucken und die BB dementsprechend ausrichten
            id= 0
            direction= Line(id, viewer, center)

            BBox.__init__(self, geoobj, direction)
            
            #Erste Seite der BB = die Seite, die direction zuerst schneidet            
            self.setfirstside(direction)
        
    
class BBgeo(BBox):

    def __init__(self, geoobj):
        """ Erzeugt eine geoglobale (an den Kartenachsen ausgerichtete) BB.
            Es muss lediglich das Polygon bergeben werden, fr das die BB
            bestimmt werden soll. Eine geoglobale BB wird fr die Berechnung
            von Himmelsrichtungen bentigt. Als erste Seite der BB wird die
            Nordseite festgelegt - das ist, mit dem Koordinatenursprung
            in der linken oberen Ecke - die Seite mit den kleinsten
            y-Koordinaten.
        """

        #Einen Richtungsvektor in Sd-Nord-Richtung erstellen
        id= 0
        direction= Line(id, Point(id,0,0), Point(id,0,1)) 
        
        #Konstruktor von BBox aufrufen
        BBox.__init__(self, geoobj, direction)
        
        #Die Nordseite zur ersten Seite der BB machen
        self.setfirstside()


    def setfirstside(self):
        """ Bestimmt die Nordseite zur ersten Seite der BB. Setzt voraus, dass
            sich der Koordinatenursprung in der linken oberen Ecke befindet.
        """

        #Ziel: die Nordseite = Seite mit den kleinsten Y-Koordinaten finden
        for zaehler in range(2):
            if self.lines[zaehler].P1.y == self.lines[zaehler].P2.y:
                #Die Nord- oder Sdseite. Aber welche davon?
                if self.lines[zaehler].P1.y < self.lines[zaehler+2].P1.y:
                    nordlinie = zaehler
                else:
                    nordlinie = zaehler + 2

        #Die Linien in die richtige Reihenfolge bringen                
        self.switchside(nordlinie)        
