# ****************************************************
# Name des Moduls: BError
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Thorsten Reichelt, Thorsten.Reichelt@web.de
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 12.01.2003
# ****************************************************

""" Dieses Modul enthlt Fehlerklassen fr die Erzeugung selbstdefinierter Fehler.
    Es gibt eine Basisklasse, BufferloError, die die allgemeine Funktionalitt im-
    plementiert. Davon abgeleitet sind Klassen fr verschiedenen Arten von Fehlern:

        GeometryError - fr Fehler bei geometrischen Berechnungen
        ParameterError - bei bergabe falscher Parameter
        SemanticError - fr semantische Fehler bei Prpositionen
        DataError - fr Fehler, die beim Lesen von Datan aus externen Quellen auftreten
        ParserError - fr Fehler beim Parsen der Prpositionalphrasen
        GUI-Error - fr GUI-Fehler
        InternalError - interne Fehler, die auf einen Bug hindeuten

    Bei der Erzeugung eines Fehlers werden immer zumindest zwei Parameter uebergeben:
    der Ort des Aufrufs (also die Funktion/ Methode, die den Fehler erzeugt hat) und
    eine Fehlernummer. Mit Hilfe der Fehlernummer kann eine genauere Fehlermeldung und
    Beschreibung aus einer Ini-Datei gelesen werden (siehe dazu BufferloError und die
    Dokumentation der Ini-Dateien). Fr das Lesen von Ini-Dateien wird die ConfigParser-
    Bibliothek importiert.
"""


import ConfigParser

class BufferloError(Exception):
    """ Basisfehlerklasse fr das Bufferlo-Projekt. Alle anderen Fehlerklassen sind von
        dieser abgeleitet. Auer __init__() werden folgende Methoden zur Verfgung gestellt:

            get_meldung(): Zum Lesen der Fehlermeldung aus der Fehler-Inidatei
            get_ursache(): Zum Lesen der nheren Fehlerbschreibung aus der Fehler-Inidatei

            __str__(): Gibt eine String-Reprsentation des Fehlers zurck. ber diese Methode
                kann die Fehlermeldung auch von Python ausgegeben werden.
    """

    def __init__(self, ort, fehler_nr):
        """ Jedem selbsterzeugten Fehler werden folgende Variablen bergeben:
                ort:        Der Ort (Name der Klasse/Funktion), wo der Fehler aufgetreten ist
                fehler_nr:  Die Nummer des Fehlers, anhand der eine Fehlerbeschreibung zugeordnet
                            werden kann
        
            Auerdem hat jedes Fehlerobjekt eine Klassenvariable name, die den Fehlertyp beschreibt
            und zum Beispiel in der Titelleiste des Fehlerfensters angezeigt werden kann.
    
            Die Fehlermeldungen und nheren Beschreibungen werden der besseren Wartbarkeit halber
            in einer Ini-Datei gespeichert, deren Name hier angegeben wird. 
        """

        self.name = "Bufferlo Fehler"
        self.inifile = "berror.ini"
        
        self.ort = ort
        self.fehler_nr = fehler_nr
        self.meldung = "Keine Meldung vorgesehen."
        self.ursache = "Keine Ursachenbeschreibung vorhanden."
 

    def get_meldung(self):
        """ Gibt in Abhngigkeit von der Fehlernummer eine Fehlermeldung zurck.
            Die Fehlermeldungen werden aus einer Ini-Datei gelesen.
        """

        meldung = ""
        config = ConfigParser.ConfigParser()
        config.read(self.inifile)

        section= self.name + "_meldungen"
        try:
            meldung = config.get(section, str(self.fehler_nr))
        except:
            meldung = "Es ist ein Fehler aufgetreten. Leider konnte aber auch die Fehlermeldung nicht aus der Datei "+self.inifile+" gelesen werden."

        return meldung            
      

    def get_ursache(self):
        """ Gibt in Abhngigkeit von der Fehlernummer eine genauere Fehlerbeschreibung zurck.
            Die Texte werden aus einer Ini-Datei gelesen.
        """

        ursache = ""
        config = ConfigParser.ConfigParser()
        config.read(self.inifile)

        section= self.name + "_ursachen"
        try:
            ursache = config.get(section, str(self.fehler_nr))
        except:
            ursache = "Bitte berprfen Sie, ob sich die genannte Datei mit den Fehlermeldungen im Programmverzeichnis befindet. Wenn ja, ist sie wahrscheinlich beschdigt und sollte ersetzt werden."

        return ursache           
            
    
    def __str__(self):
        s = "Es ist ein" + self.name + " Nr. " + str(self.fehler_nr) + " im Modul " + self.ort + " aufgetreten."
        s += "\n\nFehlermeldung:\n"+ self.meldung
        s += "\n\nUrsache / nhere Beschreibung:\n" + self.ursache
        return s
        

class GeometryError(BufferloError):
    """ Diese Klasse wird fr allerlei Fehler verwendet, die bei den geometrischen
        Berechnungen auftreten knnen - z.B., wenn mal wieder kein Lot gefllt
        werden konnte ;-)

        Anhand der Fehlernummer werden ein Fehlermeldung und eine nhere Beschreibung
        (mit Vermutungen ber die Fehlerursache) zugeordnet, die dann ausgegeben
        werden knnen.
    """

    def __init__(self, ort, fehler_nr):
        BufferloError.__init__(self, ort, fehler_nr)
        self.name = "Geometriefehler"

        self.meldung = self.get_meldung()
        self.ursache = self.get_ursache()


class ParameterError(BufferloError):
    """ Diese Klasse wird fr Fehler verwendet, die durch falsche oder fehlende Para-
        meterbergaben entstehen.

        Anhand der Fehlernummer werden ein Fehlermeldung und eine nhere Beschreibung
        (mit Vermutungen ber die Fehlerursache) zugeordnet, die dann ausgegeben
        werden knnen.
    """

    def __init__(self, ort, fehler_nr):
        BufferloError.__init__(self, ort, fehler_nr)
        self.name = "Parameterfehler"

        self.meldung = self.get_meldung()
        self.ursache = self.get_ursache()


class SemanticError(BufferloError):
    """ Diese Klasse ist fr semantische/ logische Fehler: etwa wenn die westliche
        Koordinate einer Karte grer ist als die stliche. 

        Anhand der Fehlernummer werden ein Fehlermeldung und eine nhere Beschreibung
        (mit Vermutungen ber die Fehlerursache) zugeordnet, die dann ausgegeben
        werden knnen.
    """

    def __init__(self, ort, fehler_nr):
        BufferloError.__init__(self, ort, fehler_nr)
        self.name = "Semantikfehler"

        self.meldung = self.get_meldung()
        self.ursache = self.get_ursache()


class DataError(BufferloError):
    """ Diese Klasse wird fr Fehler verwendet, die beim Lesen von Daten aus externen
        Quellen auftreten.

        Zustzlich zu den Parametern der Basisklasse erhlt die Klasse noch den Namen
        der Datenquelle, bei deren Bearbeitung der Fehler auftrat; sowie eine Beschreibung
        der Daten, deren Lesen zum Fehler fhrte.
        
        Anhand der Fehlernummer werden ein Fehlermeldung und eine nhere Beschreibung
        (mit Vermutungen ber die Fehlerursache) zugeordnet, die dann ausgegeben
        werden knnen.
    """

    def __init__(self, ort, fehler_nr, data_source, requested_data):
        BufferloError.__init__(self, ort, fehler_nr)
        self.name = "Datenfehler"
        self.data_source = data_source
        self.requested_data = requested_data

        self.meldung = self.get_meldung()
        self.ursache = self.get_ursache()


class InternalError(BufferloError):
    """ Diese Klasse ist fr interne Fehler, die auf einen Bug hindeuten.
        
        Anhand der Fehlernummer werden ein Fehlermeldung und eine nhere Beschreibung
        (mit Vermutungen ber die Fehlerursache) zugeordnet, die dann ausgegeben
        werden knnen.
    """

    def __init__(self, ort, fehler_nr):
        BufferloError.__init__(self, ort, fehler_nr)
        self.name = "Interner Fehler"

        self.meldung = self.get_meldung()
        self.ursache = self.get_ursache()


class ParserError(BufferloError):
    """ Diese Klasse ist fr Fehler, die im Parser auftreten knnen.
        
        Anhand der Fehlernummer werden ein Fehlermeldung und eine nhere Beschreibung
        (mit Vermutungen ber die Fehlerursache) zugeordnet, die dann ausgegeben
        werden knnen.
    """

    def __init__(self, ort, fehler_nr):
        BufferloError.__init__(self, ort, fehler_nr)
        self.name = "Parserfehler"

        self.meldung = self.get_meldung()
        self.ursache = self.get_ursache()


class GuiError(BufferloError):
    """ Diese Klasse ist fr Fehler, die im GUI auftreten knnen.
        
        Anhand der Fehlernummer werden ein Fehlermeldung und eine nhere Beschreibung
        (mit Vermutungen ber die Fehlerursache) zugeordnet, die dann ausgegeben
        werden knnen.
    """

    def __init__(self, ort, fehler_nr):
        BufferloError.__init__(self, ort, fehler_nr)
        self.name = "GUI-Fehler"

        self.meldung = self.get_meldung()
        self.ursache = self.get_ursache()
