# ****************************************************
# Name des Moduls: GeoObject
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Thorsten Reichelt, Thorsten.Reichelt@web.de
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 15.01.2003
# ****************************************************


""" Dieses Modul ist das Herzstck des Bufferlo-Projekts. Hier werden die
    Geltungsbereiche (Buffer) der Prpositionen berechnet. Das Modul besteht
    aus der Klasse Bufferlo und einer __main__()-Funktion, die eine Verwendung
    des Moduls ber Kommandozeilenparameter ermglicht.

    Zum Berechnen eines Buffers gengt es, eine Bufferlo-Objektinstanz zu erzeugen.
    Die Ergebnisse der Berechnung befinden sich anschlieend in self.buffer (fr
    Details siehe die Dokumentation zu Bufferlo.__init__()). Die Bedienung ber
    Kommandozeile ist in __main__ durch einen ausfhrlichen Hilfetext beschrieben.

    Folgende Module werden verwendet:
    geomap und prepcontainer: zum Laden der bentigten Objekte und Prpositions-
        daten
    geoobject: Zum Erstellen der Buffer (sind gleichfalls vom Typ GeoObject)
    bbox: Zum Berechnen von Bounding Boxes
    geometry: Fr allgemeine geometrische Berechnungen
"""


from geomap import GeoMap
from geoobject import *
from prepcontainer import PrepContainer
from bbox import *
from geometry import *
from btools import *
from berror import ParameterError, SemanticError, GeometryError, DataError, InternalError
import copy
import sys

class Bufferlo:

    def __init__(self, obj_ids, o_source, prep, p_source, viewer, mod_ids, o_type="XML", p_type="XML"):
        """ Folgende Parameter mssen bergeben werden, um einen Buffer zu berechnen:

                obj_ids: eine Liste von IDs geometrischer Objekte

                o_source: gibt den Ort der Objektdatenbank an (als String)

                prep: die ID der Prposition, fr die der/die Buffer an object
                    berechnet werden soll

                p_source: gibt den Ort der Prpositionsdatenbank an

                viewer: vom Typ ViewerObject (in geoobject) oder Point (geometry). Gibt
                    fr deiktische Sicht den Punkt an, von dem aus das Objekt gesehen
                    wird. Muss "None" sein fr alle anderen Sichtweisen.

                mod_ids: eine Liste von IDs von Modifikatoren. Diese Angabe wird im
                    Augenblick noch ignoriert, da noch keine Modifikatoren berechnet
                    werden

                o_type und p_type: Geben den Typ der Objekt- und der Prpositions-DB
                    an. Voreinstellung ist XML.

            Die Ergebnisse der Berechnungen befinden sich anschlieend in self.buffer.
            Das ist eine dreielementige Liste mit folgendem Aufbau:
                1: eine Liste der berechneten Buffer (vom Typ GeoObject)
                2: die sichtweisenabhngige BoundingBox, die bei der Berechnung ver-
                        wendet wurde
                3: die sichtweisenunabhngige BB, die bei der Berechnung verwendet wurde
            Die beiden BoundingBoxen werden mit zurckgegeben, um anschlieend von der
            GUI (siehe bgui) angezeigt werden zu knnen.

            In der Methode werden zuerst die bentigten Daten anhand der IDs geladen.
            Dann wird fr jeden Satz von Bufferparametern (siehe PrepBuffer in
            preposition), den die Prposition besitzt, ein Buffer berechnet und in
            die Ergebnisliste eingefgt                                
        """            

        #Die bentigten Objekte laden
        map= GeoMap(o_source, o_type)
        if not map.getparams():
            raise DataError("Bufferlo.init", 2, o_source, "")
        for x in obj_ids:
            if not map.getobjects(id= x):
                raise DataError("Bufferlo.init", 4, o_source, "Objekt mit der ID: " + x)

        #Prposition laden
        pcon= PrepContainer(p_source, p_type)
        if not pcon.getpreps(id= prep, projection= map.pov):
            raise DataError("Bufferlo.init", 3, p_source, "Prposition mit ID: "+prep)
        prep= pcon.cpreps[0]
        
        #Modifikatoren laden... (noch nicht untersttzt)
        mods= []

        #Fr alle Bufferparameter-Stze der Prposition einen Buffer berechnen
        self.buffer= [[]]
        for bufferparams in prep.prepbuffer:
            calc_buffer = self.calculateBuffer(bufferparams, prep.povs, map.cobjects, viewer, mods)
            self.buffer[0].append(calc_buffer[0])
        self.buffer.append(calc_buffer[1])
        self.buffer.append(calc_buffer[2])
        
        
    def calculateBuffer(self, params, povs, objects, viewer, mods):
        """ Je nach Buffertyp (zur bersicht ber die verschiedenen
            Prpositionstypen siehe die allgemeine Dokumentation) wird ein
            anderer Algorithmus zur Berechnung der Buffer verwendet. Die
            Verzweigung nach den einzelnen Typen geschieht in dieser Methode.
            Ist der Typ des Buffers unbekannt, wird ein Fehler erzeugt.
        """


        #Je nach Typ der Flche bentigen wir andere Berechnungen
        if str_equals(params.flaechentyp, "A"):
            buffer= self.calculateTypA(params, povs, objects, viewer, mods)
        else:
            raise ParameterError("Bufferlo.calculateBuffer", 2)

        return buffer        

                    
    def calculateTypA(self, params, povs, objects, viewer, mods):
        """ Diese Methode berechnet Buffer des Typ A (bersicht Buffertypen siehe
            die allgemeine Dokumentation). Fr eine genaue Beschreibung des
            Algorithmus siehe die Beschreibung der Bufferalgorithmen. Die Methode
            ist auerdem relativ ppig kommentiert ;-).
        """

        buffer = None

        #Der Typ-A-Algorithmus arbeitet nur an einzelnen Objekten.  Wenn mehrere
        #bergeben wurden, wird ein Fehler erzeugt
        if len(objects) == 1:
            obj = objects[0]
        else:
            raise InternalError("Bufferlo.calculateTypA", 2)

        #Zuerst einmal die richtige Sichweise bestimmen und die entsprechenden BoundingBoxen berechnen
        self.povmodus = None
        if viewer and listcontains(povs, "deictic"):
            bb = BBdeictic(obj, viewer)
            self.povmodus = "deictic"
            #Default-BB fr Relevanzprfung
            default_bb = BBox(obj)
            center = default_bb.getcenter()
            richtung = Line(None, Point(None, viewer.x, viewer.y), center)
            default_bb.setfirstside(richtung)
        elif listcontains(povs, "geo-global"):
            #Raumausdrcke, die geo-global sind, knnen NUR geo-global sein
            bb = BBgeo(obj)
            default_bb = BBgeo(obj)
            self.povmodus = "geo-global" 
        elif obj.direction and listcontains(povs, "intrinsic"):
            #Die Richtung eines Objekts lsst sich als Pfad, bestehend aus
            #mehreren Linien, angeben. Davon wird jedoch im Augenblick immer
            #nur die erste verwendet.
            bb = BBintrinsic(obj, obj.direction.lines[0])
            default_bb = BBintrinsic(obj, obj.direction.lines[0])
            default_bb.setfirstside(obj.direction.lines[0])
            self.povmodus = "intrinsic"
        elif listcontains(povs, "default"):
            bb= BBox(obj)
            default_bb = BBox(obj)
            self.povmodus = "default"

        if not self.povmodus:
            raise SemanticError("Bufferlo.calculateTypA", 2)
       
        #Die richtige BB-Seite bestimmen
        bbseite = params.bb_seite
        if self.povmodus == "deictic":
            bbseite = self.switchleftright(bbseite)
    
        #Die Seite der BB, an der der Buffer entstehen soll:
        bbside = bb.lines[bbseite]

        #Die Seite der default-BB, die fr die Relevanzprfung benutzt werden soll
        default_bbside = default_bb.lines[bbseite]
        
        #Die maximale Distanz berechnen, die ein Punkt zu bbside haben darf,
        #um fr die Bufferberechnung relevant zu sein
        index = indexcycle(bbseite-1, 4)
        laenge = default_bb.lines[index].length()
        max_distance = laenge * params.relevant_obj_depth / 100

        #Den Abstand berechnen, den die Bufferpunkte vom Objekt haben sollen
        inner_distance = params.inner_distance * laenge / 100
        outer_distance = params.outer_distance * laenge / 100
        
        #Die fr die Bufferprojektion relevanten Punkte ermitteln
        relevant_points = []
        relevant_lote = []
        relevant_points, relevant_lote = self.get_relevant_points_TypA(obj, bbside, default_bbside, bbseite, viewer, max_distance)

        if relevant_points:
            #Die Punkte in die richtige Reihenfolge bringen...
            relevant_points, relevant_lote = self.sort_relevant_points(relevant_points, relevant_lote, bb, bbseite)

            #...und den Buffer generieren
            buffer = self.generate_Buffer(relevant_points, relevant_lote, bb, bbseite, inner_distance, outer_distance, params.angle)

        else:
            #Das sollte eigentlich nie passieren ;-)
            raise GeometryError("Bufferlo.calculateTypA", 2)

        return [buffer, bb, default_bb]
    

    def get_relevant_points_TypA(self, obj, bbside, default_bbside, nr_bbside, viewer, max_distance):
        """ Typ-A-Algorithmus: Bestimmt die Punkte des Objekts, von denen aus der
            Buffer projeziert werden soll. Gibt zwei Listen zurck:
                1: die als relevant bestimmten Punkte
                2: die Lote dieser Punkte auf die sichweisenabhngige BB 
                        (werden fr die weitere Berechnung noch bentigt)

            Ein Objektpunkt wird als relevant fr die Buffererzeugung angesehen,
            wenn er entweder...
                a) auf der relevanten BoundingBox-Seite liegt (mindestens einer tut
                das immer, denn es ist ja eine "Bounding"Box.
            ...oder...
                b) nicht weiter als max_distance von der relevanten BB-Seite entfernt
                UND von dieser aus sichtbar ist. Fr die Sichtbarkeitsprfung siehe
                die Methode self.checkVisibility().
            Wenn auf diese Weise nur ein Punkt gefunden werden konnte (nmlich der,
            der direkt auf der BB-Seite liegt), wird versucht, von diesem Punkt
            ausgehend noch einen oder maximal zwei weitere zu finden - mit einem
            Punkt allein kann kein Buffer generiert werden. Fr Details hierzu siehe
            den dortigen Kommentar.                                    
        """
        
        #In dieser Liste werden alle relevanten Punkte des Objekts gespeichert
        relevant_points= []
        #Liste der Lote der relevanten Objektpunkte auf default_bbside
        relevant_lote= []
                
        #berprfung jedes Punktes des Objektes auf seine Buffertauglichkeit
        for p in obj.points:

            #Erstmal ne Kopie des Objektpunktes machen
            point = copy.deepcopy(p)                        

            #Falls der Punkt auf der BB-Linie ist, ist alles klar
            if collinear(bbside.P1, bbside.P2, point):
                relevant_points.append(copy.deepcopy(p))
                relevant_lote.append(None)
                
            else:
                # 2 Tests werden mit dem Punkt gemacht:
                #   - liegt er nahe genug an der BB-Seite?
                #   - ist er von der BB-Seite aus sichtbar?

                #Bentig wird das Lot vom untersuchten Punkt auf die relevante BB-Seite
                default_lot = createLot(point, default_bbside)
                if not default_lot:
                    raise GeometryError("Bufferlo.get_relevant_points_TypA:default_lot", 3)

                relevant_lot = createLot(point, bbside)
                if not relevant_lot:
                    raise GeometryError("Bufferlo.get_relevant_points_TypA:relevant_lot", 3)
                                                                
                # Falls der Punkt in Relevanzreichweite liegt und sichtbar ist...
                # ... dann Kopie des Originalobjektpunktes in Punktliste
                if (default_lot.length() <= max_distance) and (self.checkVisibility(obj, relevant_lot, viewer, point, nr_bbside)):
                    relevant_points.append(copy.deepcopy(p))
                    relevant_lote.append(copy.deepcopy(relevant_lot))
                    
        #Wurden gengend Punkte gefunden, um eine Flche zu generieren? Wenn nur ein Punkt gefunden wurde,
        #wird berprft, ob die Endpunkte der beiden Linien, in denen dieser Punkt vorkommt, sichtbar sind.
        #Ist das der Fall (und mindestens einer ist IMMER sichtbar), wird der jeweilige Punkt mit
        #aufgenommen.
        if len(relevant_points) == 1:
            punkt = copy.deepcopy(relevant_points[0])            
            for i in obj.lines:
                if punkt == i.P1:
                    lot = createLot(i.P2, bbside)
                    if not lot:
                        raise GeometryError("Bufferlo.get_relevant_points_TypA:lot3", 3)
                    if self.checkVisibility(obj, lot, viewer, i.P2, nr_bbside):
                        relevant_points.append(i.P2)
                        relevant_lote.append(copy.deepcopy(lot))
                elif punkt == i.P2:
                    lot = createLot(i.P1, bbside)
                    if not lot:
                        raise GeometryError("Bufferlo.get_relevant_points_TypA:lot4", 3)
                    if self.checkVisibility(obj, lot, viewer, i.P1, nr_bbside):
                        relevant_points.insert(0, i.P1)
                        relevant_lote.insert(0, copy.deepcopy(lot))
                                    
        return relevant_points, relevant_lote


    def sort_relevant_points(self, relevant_points, relevant_lote, bb, nr_bbside):
        """ Die relevanten Punkten sind nicht unbedingt in der richtigen Reihenfolge
            gefunden worden. Der erste Punkt in der "richtigen" Reihenfolge muss
            nmlich einer der (zwei) Punkte sein, an dem die Innenlinie des Buffers
            mit der Aussenlinie des Buffers verbunden wird. Hier werden die Punkte
            in eine dementsprechende Reihenfolge gebracht
        """
    
        position=0
        index = indexcycle(nr_bbside-1, 4)
        min_d = DPointLine(relevant_points[position], bb.lines[index])
        for x in range(len(relevant_points)):
            d = DPointLine(relevant_points[x],bb.lines[index])
            if d < min_d:
                min_d = d
                position = x
        
        relevant_points = listrotate(relevant_points, position)
        relevant_lote = listrotate(relevant_lote, position)

        return relevant_points, relevant_lote
        

    def generate_Buffer(self, relevant_points, relevant_lote, bb, nr_bbside, inner_distance, outer_distance, angle):
        """ Hier wird auf Basis der Punktliste in relevant_points das Bufferpolygon
            generiert und zurckgegeben. Das zurckgegebene Objekt ist vom Typ
            GeoObject (siehe geoobject). Fr die genaue Funktionsweise siehe die
            Kommentare im Quelltext.
        """
        
        #Zwei Listen: eine fr die innere, eine fr die uere Reihe von Punkten
        l1= []
        l2= []
        
        for x in range(len(relevant_points)):
            #Die jeweiligen inneren und ueren Punkte der Bufferflche
            #werden bestimmt, indem das Lot des Punktes, von dem die
            #beiden projiziert werden, auf die richtige Lnge gestreckt
            #wird. Dann wird der Endpunkt des Lots als Bufferpunkt gemerkt.
            if relevant_lote[x]:

                inner_lot = copy.deepcopy(relevant_lote[x])
                inner_lot.scaleToLength(inner_distance)
                l1.append(copy.deepcopy(inner_lot.P2))

                outer_lot = copy.deepcopy(relevant_lote[x])
                outer_lot.scaleToLength(outer_distance)
                l2.append(copy.deepcopy(outer_lot.P2))
                
            else:
                #Wenn es noch kein Lot gibt, liegt das daran, dass der zu projezierende Punkt
                #direkt auf der Bounding-Box-Linie liegt. In diesem Fall muss ein Lot zur
                #gegenberliegenden BB-Seite gefllt werden, da sich ein Lot mit der Lnge 0
                #nicht strecken lsst.
                
                index = indexcycle(nr_bbside + 2, 4)
                lot = createLot(relevant_points[x], bb.lines[index])
                if lot:
                    lot.reverse()
                    d = lot.length()
                else:
                    raise GeometryError("Bufferlo.generate_buffer", 3)
                
                #Gewnschte Lnge des Lots berechnen
                d += inner_distance
                l= copy.deepcopy(lot)
                l.scaleToLength(d)
                l1.append(copy.deepcopy(l.P2))

                d = d + outer_distance - inner_distance
                lot.scaleToLength(d)
                l2.append(copy.deepcopy(lot.P2))
                
        #Winkel an die beiden ueren Linien des Buffers anlegen        
        if angle:
            ende= len(l1)-1
            
            #Die linke Seite (Objekte drehen im Uhrzeigersinn):
            ln= Line(None, l1[0], l2[0])
            ln= createAngle(ln, angle)
            l2[0:0]= [ln.P2]
                        
            #Die rechte Seite:
            ln= Line(None, l1[ende], l2[ende+1])
            ln= createAngle(ln, 360 - angle)
            l2.append(ln.P2)
            
        #Jetzt noch die beiden Listen zusammenfgen        
        for x in l2[:]:
            l1.append(l2.pop())
        
        return GeoObject("Buffer", "01", l1, [])


    def checkVisibility(self, obj, lot, viewer=None, point=None, nr_bbside=None):
        """ Die Aufgabe dieser Funktion ist es, festzustellen, ob ein Objekt-
            punkt von einer BB-Seite oder dem Viewer aus sichtbar ist. Die
            Sichtbarkeitsprfung vom Viewer aus wird nur dann vorgenommen,
            wenn nr_bbside = 0 ist, es sich also um die Vorderseite des Objekts
            handelt, die der Viewer auch tatschlich sehen kann.

            Sichtbarkeitsprfung ohne Viewer:
            Es werden das geometrische Objekt (obj) und das Lot von dem fraglichen
            Punkt auf die BB-Seite bergeben. Wenn das Lot das Objekt zu oft
            schneidet, ist es von der BB-Seite aus nicht sichtbar.

            Sichtbarkeitsprfung mit Viewer:
            wie oben, nur wird das Lot vom Viewer (viewer) auf den Objektpunkt (point)
            gefllt.

            Die Funktion gibt wahr zurck, wenn der Punkt sichtbar ist, sonst falsch.            
        """

        #Wenn es um die Vorderseite geht, kann direkt vom Viewer aus gesehen werden
        if ((str_equals(self.povmodus, "deictic")) and (nr_bbside == 0)):
            line = Line(None, viewer, point)
        else:
            line = lot

        #Die Linie vom zu berprfenden Objektpunkt zum Betrachter/zur BB
        #schneidet das Objekt auf jeden Fall 2-mal, da beide Objektlinien, die
        #mit dem Objektpunkt gebildet werden, sich ja mit der Linie schneiden.
        if intersects(line, obj) < 3:
            result = 1
        else:
            result = 0 

        return result


    def switchleftright(self, bbseite):
        """ Die Nummerierung der BB-Seiten erfolgt aus intrinsischer Sicht im Uhrzeigersinn.
            Damit das auch aus deiktischer Sicht funktioniert, mssen links und rechts ver-
            tauscht werden.
        """

        if bbseite == 1:
            bbseite = 3
        elif bbseite == 3:
            bbseite = 1
        return bbseite
            


if __name__=="__main__":
    """ Bufferlo kann auch ber Kommandozeilenparameter bedient werden, um
        das Einbinden in andere Programme zu erleichtern. Wenn eine gltige
        Anzahl von Parametern eingegeben wurde, werden diese ausgewertet,
        ansonsten wird eine Hilfetext ausgegeben und 1 fr Fehler an die auf-
        rufende Instanz zurckgegeben.

        War das Berechnen der Buffer erfolgreich, werden deren Kooordinaten
        in folgender Klammerstruktur ausgegeben:
        ((Punkt, Punkt, Punkt, Punkt), (Buffer2...), (Buffer3...))
        Ist die Berechnung fehlgeschlagen, wird eine Fehlermeldung ausgegeben
        und 1 fr Fehler an die aufrufende Instanz zurckgegeben (0 bei Erfolg).
    """
    
    param_anzahl = len(sys.argv)
    if param_anzahl > 4 and param_anzahl < 11 :
        #Der erste Parameter ist immer der Name des Programms, daher stehen
        #die tatschlich bergebenen Parameter erst ab Position 1

        #Vier Parameter hats auf jeden Fall
        objects = sys.argv[1].split(",")
        obj_path = sys.argv[2]
        prep_id = sys.argv[3]
        prep_path = sys.argv[4]

        #Die restlichen Parameter mssen wir einzeln abklappern
        viewer = None
        mod_ids = None
        objdb_type = "XML"
        prepdb_type = "XML"

        #Der Viewer
        if param_anzahl > 5 and sys.argv[5] != "" :
            v_xy = sys.argv[5].split(",")
            if len(v_xy) == 2:

                #Falls der Viewer auch noch eine Blickrichtung hat
                vd_xy = None
                if param_anzahl > 6 and sys.argv[6] != "":            
                    vd_xy = sys.argv[6].split(",")
                    if len(vd_xy) != 2:
                        #Blickrichtung falsch angegeben
                        raise ParameterError("Bufferlo.main", 4)
                    
                if vd_xy:
                    viewer = ViewerObject("Viewer1", "01", v_xy, vd_xy)
                else:
                    viewer = ViewerObject("Viewer1", "01", v_xy)
            
            else:
                #Koordinaten falsch angegeben
                raise ParameterError("Bufferlo.main", 3)

        #Modifikatoren
        if param_anzahl > 7:
            mod_ids = sys.argv[7].split(",")

        #Datentypen            
        if param_anzahl > 8:
            objdb_type = sys.argv[8]
        if param_anzahl > 9:
            prepdb_type = sys.argv[9]
                
        #
        #Jetzt knnen wir versuchen, aus diesen Angaben einen Buffer zu berechnen
        #

        fehler = 0
        try:
            b = Bufferlo(objects, obj_path, prep_id, prep_path, viewer, mod_ids, objdb_type, prepdb_type)
        except:
            print "Das Berechnen des Buffers ist fehlgeschlagen. Bitte Parameter ueberpruefen!"
            fehler = 1

        #Wenn die Berechnung erfolgreich war, Buffer ausgeben
        if not fehler:
            ausgabe = "("
            for buf in b[0].buffer:
                s = "("
                for p in buf.points:
                    s += str(p.getXY())
                s += ")"
                ausgabe += s
            ausgabe += ")"
            print ausgabe

        sys.exit(fehler)            
                

    elif param_anzahl == 2:
        if sys.argv[1] == "DEBUG":
            v= ViewerObject("Oskar", "01", [1108.0, 314.0], [11.0, 10.0])
            test = Bufferlo(["11"],r"c:\bufferlo\Karten\Stadt.xml","35",r"c:\bufferlo\lingdata\praeps.xml", v, None)
        
    else:
        print """Bufferlo kann auch von der Kommandozeile aus verwendet werden.
Dazu muessen folgende Parameter angegeben werden:

    Obligatorisch:
    - eine oder mehrere Objekt-IDs (wenn mehrere, in Anfuehrungszeichen,
      mit Kommata getrennt)
    - Pfad und Name der Karte, auf der diese Objekte liegen
    - die ID der Praeposition, fuer die der Buffer berechnet werden soll
    - Pfad und Name der Praepositionsdatenbank
    
    Optional:
    - die Koordinaten des Viewers (x und y, in Anfuehrungszeichen, durch Komma
      getrennt)
    - die Blickrichtung des Viewers (genauso)
    - eine oder mehrere IDs von Modifikatoren (wenn mehrere, in Anfuehrungs-
      zeichen, durch Kommata getrennt)
    
    - Datentyp der Karte, sofern nicht XML
    - Datentyp der Praepositionsdatenbank, sofern nicht XML
    
Vom letzten Parameter ausgehend koennen nicht benoetigte Parameter wegge-
lassen werden. Soll ein Parameter zwischen anderen ausgelassen werden,
sind leere Anfuehrungszeichen anzugeben."""

        sys.exit(1)
 