# ****************************************************
# Name des Moduls: GeoMap
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 10.01.2003
# ****************************************************


""" Die Klasse GeoMap speichert eine Karte fr Bufferlo. Eine solche Karte besteht aus
    folgenden Daten (nach dem Doppelpunkt der Name der entspr. Instanzvariable):
    
        einem Namen: name
        einer ID: id
        einer Beschreibung: description
        einem Point of View - Grundriss oder Seitenansicht: pov
        4 Maximalkoordinaten fr die verschiedenen Himmelsrichtungen: maxeast, maxsouth...
        einer Liste von enthaltenen geometrischen Objekten (z.B vom Typ GeoObject, s.d.): cobjects

    Verwendung der Klasse:
    Zum Erstellen einer Karte: siehe __init__()
    Zum Setzen/ ndern von Kartenparametern: siehe setvalues()
    Zum Laden von Kartenparametern aus externen Quellen: siehe getparams()
    Zum Laden von Kartenobjekten aus externen Quellen: siehe getobjects()
    Breite und Hhe eine Karte: siehe width() und heigth()

    Zustzlich ist eine __repr__()-Funktion implementiert, so dass das Kartenobjekt per print
    ausgegeben werden kann.
          
    Es werden folgende Bibliotheken bentigt:
    maploader: enthlt die Klasse Maploader, die zum Laden von Karten aus externen Quellen benutzt wird
    berror: fr selbstdefinierte Exceptions
"""


from maploader import Maploader
from berror import ParameterError, SemanticError


class GeoMap:
    
    def __init__(self, sourcename= "", sourcetyp= ""):
        """ Legt eine leere Liste fr enthaltene Objekte an und initialisiert die
            brigen Kartenparameter. Um den Kartenparametern Werte zuzuweisen, kann
            die Funktion setvalues() benutzt werden.
            
            Wenn sourcename und sourcetyp angegeben wurden, wird versucht, ein Loaderobjekt
            anzulegen. Wenn das erfolgreich ist (die Klasse Maploader z.B. den angegebenen
            sourcetyp untersttzt (siehe da), knnen anschlieend Kartenparameter und Objekte
            aus einer externen Quelle names sourcename geladen werden. Hierzu siehe getparams()
            und getobjects().
        """
                       
        self.cobjects= []
        self.loader = None
        
        self.name= ""
        self.description= ""
        self.id= 0
        self.pov= ""

        self.maxwest= 0
        self.maxeast= 0
        self.maxnorth= 0
        self.maxsouth= 0

        if sourcename and sourcetyp:
            self.loader= Maploader(self, sourcename, sourcetyp)
            
    
    def setvalues(self, name, id, pov, maxwest, maxeast, maxnorth, maxsouth, description=""):
        """ Initialisiert eine Karte mit den bergebenen Werten und prft die 
            Kartengrenzen auf Plausibilitt.
        """
        
        self.name= name
        self.description= description
        self.id= id
        self.pov= pov

        self.maxwest= int(maxwest)
        self.maxeast= int(maxeast)
        self.maxnorth= int(maxnorth)
        self.maxsouth= int(maxsouth)

        #Plausibilittsprfungen
        if self.id== "": raise ParameterError("Geomap.setvalues", 5)
        if ((self.maxeast <= self.maxwest) or (self.maxsouth <= self.maxnorth)):
            raise SemanticError("Geomap.setvalues", 3)

        
    def height(self):
        """ Gibt die Hhe der Karte zurck. """
        return self.maxsouth - self.maxnorth


    def width(self):
        """ Gibt die Breite der Karte zurck """
        return self.maxeast - self.maxwest


    def addobject(self, newobj):
        """ Fgt ein Objekt in die Objektliste der Karte ein. berprft dabei, ob
            die ID des neuen Objekts auf der Karte schon vorkommt und erzeugt einen
            Fehler, falls das der Fall ist.
        """

        contained = 0
        for obj in self.cobjects:        
            if obj.id == newobj.id:
                contained = 1
                break

        if not contained:
            self.cobjects.append(newobj)
        else:
            #Fehler: Ein Objekt mit dieser ID gibt es schon
            raise ParameterError("Geomap.addobject", 7)
            
    
    def getparams(self):
        """ Liest die Kartenparameter aus einer externen Quelle, benutzt dazu das Maploader-Objekt
            (zur Definition einer externen Quelle siehe __init__()).
            
            Es wird 1 zurckgegeben, wenn erfolgreich Daten gelesen wurden, 0 bei Misserfolg.
            War das Laden erfolgreich, sind die Kartenparameter anschlieend dementsprechend gesetzt.
        """
            
        success= 0

        if self.loader:
            success= self.loader.getparams()

        return success            


    def getobjects(self, name= "", id= ""):
        """ Liest die geometrischen Objekte der Karte aus einer externen Quelle, benutzt dazu
            das Maploader-Objekt(zur Definition einer externen Quelle siehe __init__().
        
            Es knnen Name und/ oder ID der geometrischen Objekte angegeben werden, die geladen 
            werden sollen. Wird nichts von beidem angegeben, werden alle Objekte geladen. In jedem Fall 
            befinden sich die geladenen Objekte anschlieend in cobjects.
            
            Es wird 1 zurckgegeben, wenn erfolgreich Daten gelesen wurden, 0 bei Misserfolg.
        """

        success= 0

        if self.loader:
            success= self.loader.getobjects(name, id)

        return success  
                
        return 1
        
        
    def __repr__(self):
        """ bergibt die gesamte Karte als String, der dann ausgegeben werden kann."""
        
        ausgabe= "Karte:"+ self.name
        ausgabe+= "\nBeschreibung:"+ self.description
        ausgabe+= "\nID:"+ self.id
        ausgabe+= "\nSichtweise:"+ self.pov
        
        ausgabe+= "\n\nWestlicher/ Oestlicher Extrempunkt: "+ str(self.maxwest)+ "/"+ str(self.maxeast)
        ausgabe+= "\nNoerdlicher/ suedlicher Extrempunkt: "+ str(self.maxnorth)+ "/"+ str(self.maxsouth)
        
        ausgabe+= "\nAuf der Karte befinden sich folgende Objekte:"
        for x in self.cobjects:
            ausgabe+= "\n"+ x.__repr__()
            
        return ausgabe
    