# ****************************************************
# Name des Moduls: GeoObject
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Thorsten Reichelt, Thorsten.Reichelt@web.de
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 10.01.2003
# ****************************************************


""" Dieses Modul enthlt zwei Klassendefinitionen. Beide Klassen sind von
    grundlegenden geometrischen Klassen aus dem Modul geometry von Titus
    von der Malsburg abgeleitet.

    Die erste Klasse ist GeoObject. In ihr knnen die Informationen eines
    geometrischen Objekts, wie sie auf den Bufferlo-Karten verwendet werden,
    gespeichert werden. Sie ist von geometry.Polygon abgeleitet. Um die
    grammatikalischen Informationen eines solchen Objekts speichern zu knnen,
    wird aus dem Modul gram noch Nounlike importiert.

    Die zweite Klasse ist ViewerObject. Eine Instanz dieser Klasse reprsentiert
    den Betrachter, der auf die Bufferlo-Karte gesetzt werden kann, um eine
    Prpositionsbufferberechnung aus deiktischer Sicht zu ermglichen. Diese
    Klasse ist von geometry.Point abgeleitet.
"""

#Auf der Basis von Titus Biblithek
from geometry import *

from gram import Nounlike


class GeoObject(Polygon):

    """ Ein GeoObject besteht aus folgenden Daten (nach dem Doppelpunkt die Namen der
        entsprechenden Instanzvariablen):
    
        einem Name: name
        einer ID: id
        eine Liste von Punkten: points (Instanzen von Point, siehe geometry.py)
        (aus diesen Punkten generiert die __init__()-Methode des Polygons auch noch eine Liste
         von Linien: lines)
    
        einer Liste von Punkten, die die Blickrichtung des Objekts angeben: dpoints (ebf. Point)
        einem Dictionary mit semantischen Werten: semantics
        einer Liste von Objekt-IDs der in diesem Objekt enthaltenen Objekte: contained
    
        Fr eine genauere Beschreibung der in diesem Objekt gespeicherten Informationen siehe
        auch die Beschreibung von odb.dtd (die DTD der Objektdatenbanken).

        Verwendung der Klasse:
        Anlegen eines neuen Objekts: __init__()
        Stringreprsentation: __repr__()
    """


    def __init__(self, name, id, points, dpoints, semantics= {}, contained= [], graminfos= None):
        """ Legt ein neues Objekt mit den bergebenen Werten an. """
        
        #Konstruktor des Polygons aufrufen
        Polygon.__init__(self, id, points)

        self.name= name
        self.id= id

        #Wenn eine Richtung fuer das Objekt angegeben wurde, als Pfad speichern
        if (dpoints != []):
            self.direction= Path(None, dpoints)
        else:
            self.direction= ""
        
        #Semantik als Dictionary, so lassen sich die Werte am einfachsten abfragen
        self.semantics= semantics

        #Liste von IDs der beinhalteten Objekte
        self.contained= contained

        #Grammatische Informationen im Nounlike-Objekt
        self.graminfos= graminfos
        

    def __repr__(self):
        """ bergibt das Objekt als String, der dann ausgegeben werden kann. """
        
        ausgabe= "Objekt: "+ self.name + " (ID: "+str(self.ID)+")\n"
        ausgabe+= "\nPunkte:\n"

        for point in self.points:
            ausgabe+= str(point.getPoints())+"\n"
        ausgabe+= "\nRichtung:\n"

        if self.direction:
            for point in self.direction.points:
                ausgabe+= str(point.getPoints())+"\n"

        ausgabe+= "\nSemantik:\n" + str(self.semantics) + "\n"
        ausgabe+= "\nEnthaltene Objekte:\n" + str(self.contained)

        if self.graminfos:
            ausgabe+= "\nGrammatische Informationen:\n"
            ausgabe += self.graminfos.__repr__()
            
        return ausgabe


class ViewerObject(Point):
    """ Die Klasse Viewerobject beinhalted die Daten des im Bufferlo-Projekt
        verwendeten Betrachters. Sie basiert auf der Klasse Point aus geometry.py.

        Eine ViewerObject-Instanz bentigt folgende Daten:

        name: Ein Name
        id: Eine ID
        viewer_coords: 2er Liste mit x und y Koordinate des Betrachters [x, y]
        direction_coords: 2er Liste mit x und y Komponente des Richtungsvektors [x, y]

        Verwendung der Klasse:
        Erstellen eines Objekts: objektinstanz = ViewerObject("Hans", "01", [344,799], [2,3])
        Stringreprsentation des Objekts: siehe __repr__                   
    """

    def __init__(self, name, id, viewer_coords = [0, 0], direction_vector = None):
        """ Legt ein neues Objekt der Klasse ViewerObject an. Daten siehe Klassen-
            beschreibung.
        """

        #Konstruktor des Point-Objekts aufrufen
        Point.__init__(self, id, float(viewer_coords[0]), float(viewer_coords[1]))

        #zustzliche Daten speichern
        self.name = name
        self.id = id

        #Falls vorhanden, Direction abspeichern
        self.direction = None
        if direction_vector:
            self.direction = [float(direction_vector[0]), float(direction_vector[1])]


    def __repr__(self):
        """ Gibt die Objektdaten als formatierter Sting zurck, der dann ausgegeben werden kann."""

        ausgabe = "\nViewer: " + self.name + " (ID: "+str(self.ID)+")\n"
        ausgabe += "Koordinaten: " + str(self.x) + "/" + str(self.y) + "\n"
        ausgabe += "Richtungsvektor   : " + str(self.direction[0]) + "/" + str(self.direction[1]) + "\n"

        return ausgabe
