# ****************************************************
# Name des Moduls: Gram
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 11.01.2003
# ****************************************************


""" Zum Analysieren von Prpositionalphrasen bentigt der Parser (siehe ptag.py)
    auer den Daten der Prpositionen (in der Prpositions-DB) und denen der Nomen
    (die Objekte auf den Karten) auch noch die Definition von Artikeln und sog.
    Blends. Fr diese Elemente werden hier Klassen definiert, auerdem gibt es eine
    Klasse Gramloader, mit der diese Wrter aus der Grammatikdatenbank gelesen werden
    (siehe auch die Dokumentation der Grammatik-DB).

    Folgende Klassen befinden sich im Modul:

    - Word: Grundklasse fr eine grammatikalische Wortdefinition.
    - Nounlike: von Word abgeleitet, mit nomenspezifischen Informationen wie
     Kasus, Numerus und Genus angereichert. In einem Nounlike-Objekt werden
      auch die grammatikalischen Infos eines Kartenobjekts gespeichert (siehe
      geoobject)
    - Determinier: von Nounlike abgeleitet. Definition eines Artikelworts.
    - Blend: speichert eine Ersetzungsregel der Form (am --> an dem)

    - Gramloader: zum Laden von Artikeln und Blends aus externen Quellen
    - Determiner_XMLHandler und Blends_XMLHandler: Handler fr den Python
      Sax-Parser zum Laden von Artikeln und Blends aus XML-Dateien.
"""


from btools import str_equals

# -------- XML -------------

#Fr den XML-Parser muss man Handler definieren, die die Daten verarbeitet -
#siehe GeoMap_Paramhandler u. GeoMap_Objecthandler
from xml.sax import handler

#Steuert den SAX-XML-Parser an. Wird zum Laden von Artikeln, Mods und Blends bentigt
from xmlparser import *


class Word:
    """ Definiert ein beliebiges Wort. Es hat einen Namen (das Wort selbst) und eine
        Zitierform bzw. Grundform, das Lexem.
    """

    def __init__(self, name, lexem= ""):
        """ Das Wort bentigt lediglich einen Namen. Die Angabe eines Lexems ist
            optional
        """
                        
        self.name= name
        if lexem:
            self.lexem= lexem
        else:
            self.lexem= self.name


class Nounlike(Word):
    """ Erweitert die Wortdefinition um nomentypische Eigenschaften wie Numerus, Kasus
        und Genus
    """
    
    def __init__(self, name, lexem, kasus, numerus, genus):
        """ Es mssen ein Name, ein Lexem und Kasus, Numerus und Genus des Wortes
            angegeben werden. Die Angabe eines Lexems ist nicht mehr optional,
            da Nomen normalerweise durch Flektion ihre morphologische Gestalt
            verndern """
    
        #Konstruktor der Basisklasse aufrufen
        Word.__init__(self, name, lexem)

        self.kasus= kasus
        self.numerus= numerus
        self.genus= genus
             
        
    def __repr__(self):
        """ Gibt das Objekt als String zurck, der dann ausgegeben werden kann """

        ausgabe = "Name: "+ self.name
        ausgabe += "\nLexem: " + self.lexem
        ausgabe += "\nKasus: " + self.kasus
        ausgabe += "\nNumerus: " + self.numerus
        ausgabe += "\nGenus: " + self.genus

        return ausgabe
    
    
class Determiner(Nounlike):
    """ Definiert ein Artikelwort. Unterscheidet sich von der Nounlike-Definition
        lediglich dadurch, dass eine Part-of-Speech-Angabe (pos) auf "Art" fest-
        gesetzt wird.
    """
    
    def __init__(self, name, lexem, kasus, numerus, genus):
                
        #Konstruktor der Basisklasse aufrufen
        Nounlike.__init__(self, name, lexem, kasus, numerus, genus)

        #Ich bin ein Artikelwort!
        self.pos = "Art"

    def __repr__(self):

        ausgabe = "Parameter des Artikelworts:\n"
        ausgabe += Nounlike.__repr__(self)
        ausgabe += "\n"
        return ausgabe

        
class Blend:
    """ Diese Klasse definiert Ersetzungsregeln zur Auflsung  morphologische Verschmelzungen
        wie "am -> an". Es wird auch eine Methode replace() angeboten, die in einem bergebenen
        String alle Vorkommen dieser Blend ersetzt.
    """

    def __init__(self, original, replacement):
        """ original enthlt das verschmolzene Wort (am), replacement die Ersetzung (an dem) """
        
        self.original= original
        self.replacement= replacement

    def replace(self, text):
        """ Ersetzt alle Vorkommen von self.original in text durch self.replacement """

        words= text.split()
        erg= ""
        for x in range(len(words)):
            if str_equals(words[x], self.original):
                words[x]= self.replacement
            erg += words[x]+ ' '

        return erg.strip()



class Gramloader:
    """ Mit Hilfe dieser Klasse knnen Grammatikdaten (Artikel und Blends) aus externen Quellen,
        im Augenblick lediglich XML-Dateien, geladen werden.
    """

    def __init__(self, sourcename, sourcetyp):
        """ Zum Erstellen eines Objekts werden folgende Parameter bentigt:

            der Name der Quelle, aus der Daten geladen werden sollen: sourcename
            der Typ der Quelle, aus der geladen werden soll: sourcetyp
        
            Auerdem werden zwei Listen angelegt, ber die man anschlieend auf
            die geladenen Blends bzw. Artikel zugreifen kann
        """
        
        self.sourcename= sourcename
        self.sourcetyp= sourcetyp.upper()

        self.blends = []
        self.artikel = []

    def getdeterminer(self, name):
        """ Ldt, je nach angegebenem sourcetyp, Artikeldefinitionen aus einer
            externen Quelle. Die geladenen Artikel stehen anschlieend in der Liste
            self.artikel. Die Funktion gibt 1 zurck, wenn erfolgreich Artikel
            geladen werden konnten, und 0 sonst.
        """

        success= 0
        self.artikel= []
        if (str_equals(self.sourcetyp, "XML")):
            #Der Sax-Handler fuer gefundene Daten
            hd= Determiner_XMLHandler(self.artikel, name)

            #Einen Parser mit unserem Handler erschaffen...
            parser= XMLParser(hd, self.sourcename)

            #... und ihn schuften lassen
            parse_success = self.xmlparse(parser)

            #Noch nachsehen, ob der Handler Schwierigkeiten hatte
            if parse_success:
                success = hd.success
            
        return success


    def getblends(self):
        """ Ldt alle Blends aus der angegebenen externen Quelle, sie befinden sich
            anschlieend in der Liste self.blends. Je nach sourcetyp wird ein
            entsprechender Parser initialisiert und aufgerufen.

            Anmerkung: entgegen den brigen get...()-Funktionen, mit denen Prps, Objekte
            etc. geladen werden knnen, wird hier nicht berprft, ob der Handler seine
            Arbeit erfolgreich beenden konnte (also eine Blend gelesen wurde). Das ist
            deshalb, da es Sprachen geben mag, in denen es kleine Blend-Phaenomene gibt.
            Trotzdem aber soll die aufrufende Funktion nicht denken, es sei ein Fehler
            aufgetreten.            
        """

        success= 0
        self.blends= []
        if (str_equals(self.sourcetyp, "XML")):
            #Der Sax-Handler fuer gefundene Daten
            hd= Blends_XMLHandler(self.blends)

            #Einen Parser mit unserem Handler erschaffen...
            parser= XMLParser(hd, self.sourcename)

            #... und ihn schuften lassen
            success= self.xmlparse(parser)
        
        return success


    def xmlparse(self, parser):
        """ Ruft den XML-Parser auf. 
            Gibt 1 bei Erfolg, 0 bei Misserfolg zurck. """

        success= 1
        try:
            parser.parse()
        except:
            success= 0

        return success            
                

class Determiner_XMLHandler(handler.ContentHandler):
    """ Diese Klasse wird vom Sax-XML-Parser benutzt. Sie soll Artikel aus einer Grammatik-DB
        im XML-Format fischen. Alle Methoden auer __init__() werden vom Sax-Parser bei
        bestimmten Ereignissen aufgerufen - zum Beispiel dem Erreichen eines schlieenden Tags.
        Ihre Namen und Schnittstellen sind also durch dessen Vorgaben festgelegt - siehe auch
        die Python-Dokumentation.
        
        Verwendung/ Aufruf der Methoden:
        Zur Erstellung eines neuen Handlers: __init__()
        Zum Verarbeiten von Inhalt (Zeichen innerhalb eines XML-Tags): characters()
        Beim Erreichen eines ffnenden Tags: startElement()
        Beim Erreichen eines schlieenden Tags: endElement()
        
        Verwendete Bibliotheken:
        Die Klasse muss von xml.sax.handler abgeleitet werden - also wird diese importiert.
    """

    def __init__(self, artikelliste, searchname):
        """ Initialisierungen, Verweis auf Artikelliste und Suchwort speichern """
                
        self.artikel= artikelliste
        self.searchname= searchname

        #Erfolg berwachen
        self.success= 0

        #Ein paar Initialisierungen
        self.inhalt= ''
        self.attention = 0

        self.initDataVars()


    def initDataVars(self):
        
        self.name= ""
        self.lexem= ""
        self.kasus= ""
        self.numerus= ""
        self.genus= ""
        
        
    def characters(self, content):
        """ Wird immer dann aufgerufen, wenn Text zwischen Tags gefunden wurde - dieser wird
            in content bergeben. Die Methode macht nichts weiter, als sich den Text in einer
            Instanzvariable inhalt zu merken - verarbeitet wird der Inhalt immer beim Erreichen
            eines schlieenden Tags - also in endElement().
        """
        
        self.inhalt+= content


    def startElement(self, name, attrs):
        """ Diese Methode wird immer aufgerufen, wenn ein ffnender Tag gefunden wurde. Hier 
            fungiert sie als eine Art Wchter. Ein Artikel beginnt mit einem gleichnamigen Tag -
            nur wenn der gefunden wird, wird attention auf 1 gesetzt. Und nur wenn attention= 1
            ist, wird in endElement(), wo die Daten verarbeitet werden, berhaupt irgendetwas getan.
        
            Dieser Methode werden vom Sax-Parser auch die Attribute der Tags bergeben - die
            hier gemerkt und ebenfalls in endElement() weiterverarbeitet werden.
        """

        if name == "Artikel":
            #Jetzt heissts aufpassen!
            self.attention= 1

        if self.attention:
            #Die Attribute merken, sie werden auch in endElement verarbeitet
            self.attrs= attrs.copy()
        else:
            self.inhalt= ""
    
    
    def endElement(self, name):
        """ Diese Methode wird immer aufgerufen, wenn ein schlieender Tag gefunden wurde. Da
            man dann davon ausgehen kann, dass eine von characters() gefundene Informationseinheit
            abgeschlossen wurde, werden die gelesenen Daten hier weiterverarbeitet, d.h., je nach
            Name des schlieenden Tags (der vom Parser ergeben wird) entsprechenden Variablen
            zugeordnet. 
            
            Wenn der Tag </Artikel> gefunden und also geschlossen wurde, wird attention auf 0 
            gesetzt - der Rest der Datei sollte jetzt am Parser vorbeirauschen. Der gelesene Artikel
            wird an die vom aufrufenden Ort bergebene Liste angehngt.
        """
        
        if self.attention:
            self.inhalt= normalize_str(self.inhalt)
            if name== "Lexem":
                self.lexem= self.inhalt

                #berprfen, ob es sich um den gesuchten Artikel handelt
                if self.lexem != self.searchname:
                    self.attention= 0
                                
            elif name== "Name":
                self.name= self.inhalt
            elif name== "Genus":
                self.genus= self.inhalt
            elif name== "Num":
                self.numerus= self.inhalt
            elif name== "Kasus":
                self.kasus= self.inhalt
            
            self.inhalt= ''

            if name== "Artikel":
                #Das wars...
                self.attention = 0
                self.success = 1

                #Die gewonnenen Daten dem Kartenobjekt bergeben
                art= Determiner(self.name, self.lexem, self.kasus, self.numerus, self.genus)
                self.artikel.append(art)
                self.initDataVars()
                
        else:
            self.inhalt= ""
            
            

class Blends_XMLHandler(handler.ContentHandler):
    """ Diese Klasse wird vom Sax-XML-Parser benutzt. Sie soll Ersetzungsregeln aus einer Grammatik-DB
        im XML-Format fischen. Alle Methoden auer __init__() werden vom Sax-Parser bei
        bestimmten Ereignissen aufgerufen - zum Beispiel dem Erreichen eines schlieenden Tags.
        Ihre Namen und Schnittstellen sind also durch dessen Vorgaben festgelegt - siehe auch
        die Python-Dokumentation.
        
        Verwendung/ Aufruf der Methoden:
        Zur Erstellung eines neuen Handlers: __init__()
        Zum Verarbeiten von Inhalt (Zeichen innerhalb eines XML-Tags): characters()
        Beim Erreichen eines ffnenden Tags: startElement()
        Beim Erreichen eines schlieenden Tags: endElement()
        
        Verwendete Bibliotheken:
        Die Klasse muss von xml.sax.handler abgeleitet werden - also wird diese importiert.
               
    """

    def __init__(self, blends):
        """ Initialisierungen, Verweis auf Blendliste speichern """
                
        self.blends= blends
        
        #Erfolg berwachen
        self.success= 0

        #Ein paar Initialisierungen
        self.inhalt= ''
        self.attention = 0

        self.initDataVars()


    def initDataVars(self):
        
        self.original= ""
        self.replacement= ""
               
        
    def characters(self, content):
        """ Wird immer dann aufgerufen, wenn Text zwischen Tags gefunden wurde - dieser wird
            in content bergeben. Die Methode macht nichts weiter, als sich den Text in einer
            Instanzvariable inhalt zu merken - verarbeitet wird der Inhalt immer beim Erreichen
            eines schlieenden Tags - also in endElement.
        """
        
        self.inhalt+= content


    def startElement(self, name, attrs):
        """ Diese Methode wird immer aufgerufen, wenn ein ffnender Tag gefunden wurde. Hier 
            fungiert sie als eine Art Wchter. Eine Regel beginnt mit einem Tag namens Blend -
            nur wenn der gefunden wird, wird attention auf 1 gesetzt. Und nur wenn attention= 1
            ist, wird in endElement(), wo die Daten verarbeitet werden, berhaupt irgendetwas getan.
        
            Dieser Methode werden vom Sax-Parser auch die Attribute der Tags bergeben - die
            hier gemerkt und ebenfalls in endElement() weiterverarbeitet werden.
        """

        if name == "Blend":
            #Jetzt heissts aufpassen!
            self.attention= 1

        if self.attention:
            #Die Attribute merken, sie werden auch in endElement verarbeitet
            self.attrs= attrs.copy()
        else:
            self.inhalt= ""
    
    
    def endElement(self, name):
        """
            Diese Methode wird immer aufgerufen, wenn ein schlieender Tag gefunden wurde. Da
            man dann davon ausgehen kann, dass eine von characters() gefundene Informationseinheit
            abgeschlossen wurde, werden die gelesenen Daten hier weiterverarbeitet, d.h., je nach
            Name des schlieenden Tags (der vom Parser ergeben wird) entsprechenden Variablen
            zugeordnet. 
            
            Wenn der Tag </Blend> gefunden und also geschlossen wurde, wird attention auf 0 
            gesetzt - der Rest der Datei sollte jetzt am Parser vorbeirauschen. Der gelesene Artikel
            wird an die vom aufrufenden Ort bergebene Liste angehngt.
        """
        
        if self.attention:
            self.inhalt= normalize_str(self.inhalt)
            if name== "String":
                self.original= self.inhalt
            elif name== "Ersatz":
                self.replacement= self.inhalt
            
            self.inhalt= ''

            if name== "Blend":
                #Das wars...
                self.attention = 0
                self.success = 1

                #Die gewonnenen Daten dem Kartenobjekt bergeben
                blend= Blend(self.original, self.replacement)
                self.blends.append(blend)
                self.initDataVars()                

        else:
            self.inhalt= ""
