# ****************************************************
# Name des Moduls: MapLoader
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 09.01.2003
# ****************************************************


""" Die Maploader-Bibliothek wird von der Klasse GeoMap verwendet, um Kartendaten aus
    externen Quellen laden zu knnen.
    
    Das Verstndnis dieser Bibliothek ist nicht ntig, um mit GeoMap-Objekten umgehen zu
    knnen. Es handelt sich um eine INTERNE Bibliothek.
    
    Die Bibliothek enthlt die Klasse Maploader. Um Daten laden zu knnen, wird von GeoMap
    ein Maploader-Objekt erstellt. Diesem werden der Name des zu verwendenden Formats und 
    der Name der Quelle (Dateiname, URL) bergeben. Beim Aufruf der Methoden getparams()
    oder getobjects() wird dann je nach Format ein entsprechender Parser aufgerufen.
    
    Im Augenblick lediglich das XML-Format untersttzt. Es wird der von Python mitgelieferte
    Sax-Parser verwendet, um die Objektdatenbanken zu lesen. Dem Parser muss ein Objekt ber-
    geben werden, dessen Methoden zur Verarbeitung der gelesenen Daten aufgerufen werden. Die
    entsprechenden Klassen heien GeoMap_XMLParamHandler (fr das Lesen der Kartenparameter) 
    und GeoMap_XMLObjectHandler (fr die geometrischen Objekte).
    
    Wahrscheinlich erfordert der Sax-Parser eine Python-Version >= 2.2 !!!
"""


from string import *
from geometry import Point
from geoobject import GeoObject
from gram import Nounlike
from berror import SemanticError, ParameterError
from btools import str_equals


# -------- XML -------------

#Fr den XML-Parser muss man Handler definieren, die die Daten verarbeitet -
#siehe GeoMap_Paramhandler u. GeoMap_Objecthandler
from xml.sax import handler

#Steuert den SAX-XML-Parser an. Wird zum Laden der Karte bentigt
from xmlparser import *


class Maploader:
    """ Mit Hilfe dieses Objekts kann ein GeoMap-Objekt Kartendaten aus externen Quellen laden.
        Dabei sollen Art und Struktur dieser Quellen fr das Kartenobjekt unsichtbar bleiben.
        
        Verwendung der Klasse:
        Erstellen eines Maploader-Objekts: siehe __init__()
        Zum Laden der Kartenparameter: siehe getparams()
        Zum Laden von Kartenobjekten: siehe getobjects()
        
        Es werden folgende Bibliotheken bentigt:
            Fr den XML-Parser: XMLParser
    """


    def __init__(self, mapobject, sourcename, sourcetyp):
        """ Zum Erstellen eines Objekts werden folgende Parameter bentigt:
        
        ein Verweis auf das Kartenobjekt, fr das Daten geladen werden sollen (also
        eine Instanz von GeoMap): mapobject
        der Name der Quelle, aus der Daten geladen werden sollen: sourcename
        der Typ der Quelle, aus der geladen werden soll: sourcetyp
        
        Diese Daten werden gespeichert, ansonsten geschieht hier nichts mehr. """
        
        self.mapobject= mapobject
        self.sourcename= sourcename
        self.sourcetyp= sourcetyp.upper()


    def getparams(self):
        """ Diese Methode wird aufgerufen, um die Kartenparameter zu laden. Das sind die all-
            gemeinen Angaben zu einer Karte, ohne die geometrischen Objekte (fr nhere Beschreibung
            der Kartenparameter siehe GeoMap bzw. die Dokumentation des Kartenformats). 
        
            Je nach sourcetyp wird ein entsprechender Parser initialisiert und aufgerufen. War das
            Laden der Parameter erfolgreich (diese werden direkt in das Kartenobjekt geschrieben),
            gibt die Methode 1 fr Erfolg zurck, 0 bei Misserfolg.
        """

        success= 0
        if (str_equals(self.sourcetyp, "XML")):
            #Der Sax-Handler fuer gefundene Daten
            hd= GeoMap_XMLParamHandler(self.mapobject)

            #Einen Parser mit unserem Handler erschaffen...
            Paramparser= XMLParser(hd, self.sourcename)

            #... und ihn schuften lassen
            parse_success= self.xmlparse(Paramparser)

            if parse_success:
                success= hd.success
            
        return success            


    def getobjects(self, name, id):
        """ Ldt die Objekte einer Karte. Es knnen ein Name und/ oder eine ID angegeben werden -
            dann werden nur die Objekte mit diesem Namen oder das Objekt mit dieser ID geladen.
            
            Je nach sourcetyp wird ein entsprechender Parser initialisiert und aufgerufen. War das
            Laden des Objekts/ der Objekte erfolgreich (diese werden direkt in das Kartenobjekt 
            geschrieben), gibt die Methode 1 fr Erfolg zurck, 0 bei Misserfolg.
        """

        success= 0
        if (str_equals(self.sourcetyp, "XML")):
            #Der Sax-Handler fuer gefundene Daten
            hd= GeoMap_XMLObjectHandler(self.mapobject, name, id)

            #Einen Parser mit unserem Handler erschaffen...
            Objectparser= XMLParser(hd, self.sourcename)

            #... und ihn schuften lassen
            parse_success= self.xmlparse(Objectparser)
    
            if parse_success:
                #Nachsehen, ob der Handler Probleme hatte
                success= hd.success

        return success


    def xmlparse(self, parser):
        """ Ruft den XML-Parser auf. Der Aufruf wird hier zentralisiert, um die Fehlerbehandlung
            nicht mehrmals schreiben zu mssen - der Parser muss nmlich sowohl in getparams() als
            auch in getobjects aufgerufen werden. Dort wird auch ein Parserobjekt mit dem richtigen
            Handler initialisiert und hier in parser bergeben.
            
            Gibt 1 bei Erfolg, 0 bei Misserfolg zurck. """

        success= 1
        try:
            parser.parse()
        except (SemanticError, ParameterError):
            raise
        except:
            success= 0
            raise

        return success            
        

class GeoMap_XMLParamHandler(handler.ContentHandler):
    """ Diese Klasse wird vom Sax-XML-Parser benutzt. Sie soll die GeoMap-Kartenparameter aus
        einer Objektdatenbank im XML-Format fischen. Alle Methoden auer __init__() werden vom
        Sax-Parser bei bestimmten Ereignissen aufgerufen - zum Beispiel dem Erreichen eines 
        schlieenden Tags. Ihre Namen und Schnittstellen sind also durch dessen Vorgaben 
        festgelegt - siehe auch die Python-Dokumentation.
        
        Verwendung/ Aufruf der Methoden:
        Zur Erstellung eines neuen Handlers: __init__()
        Zum Verarbeiten von Inhalt (Zeichen innerhalb eines XML-Tags): characters()
        Beim Erreichen eines ffnenden Tags: startElement()
        Beim Erreichen eines schlieenden Tags: endElement()
        
        Verwendete Bibliotheken:
        Die Klasse muss von xml.sax.handler abgeleitet werden - also wird diese importiert.
        
        Die dieser Klasse erwartete XML-Struktur ist in odb.dtd beschrieben. Siehe dort und
        die dafr vorgesehene Dokumentation
    """

    def __init__(self, mapobject):
        """ Beim Erstellen eines Handlerobjekts muss ein Verweis auf das Kartenobjekt ber-
            geben werden, fr das die Daten gelesen werden - nur so knnen sie nmlich auch
            gleich dort gespeichert werden. Aber diesen Verweis hat ja schon das Maploader-
            Objekt bekommen, und er wird nach hier durchgereicht (siehe die dortige Init-Methode).
            
            Ansonsten geschieht nichts weiter - ein paar Variablen werden initialisiert (unter
            anderem eine Instanzvariable success, an der Maploader feststellen kann, ob die
            Ausfhrung erfolgreich war. Anfangswert = 0 = Misserfolg)
        """
        
        
        self.mapobject= mapobject

        #Erfolg berwachen
        self.success= 0

        #Ein paar Initialisierungen
        self.inhalt= ''
        self.attention = 0
        
        #Kartenparameter
        self.mapname= ""
        self.pov= ""
        self.id= ""
        self.description= ""
        self.maxwest= ""
        self.maxeast= ""
        self.maxnorth= ""
        self.maxsouth= ""

        
    def characters(self, content):
        """ Wird immer dann aufgerufen, wenn Text zwischen Tags gefunden wurde - dieser wird
            in content bergeben. Die Methode macht nichts weiter, als sich den Text in einer
            Instanzvariable inhalt zu merken - verarbeitet wird der Inhalt immer beim Erreichen
            eines schlieenden Tags - also in endElement.
        """
        
        self.inhalt += content


    def startElement(self, name, attrs):
        """ Diese Methode wird immer aufgerufen, wenn ein ffnender Tag gefunden wurde. Hier 
            fungiert sie als eine Art Wchter. Alle Kartenparameter stehen zwischen den
            <mapinfo>-Tags. Wenn also ein ffnendes <mapinfo> gefunden wird, wird es interessant,
            ein Flag namens attention wird auf 1 gesetzt. Und nur wenn attention= 1 ist, wird
            in endElement(), wo die Daten verarbeitet werden, berhaupt irgendetwas getan.
        
            Dieser Methode werden vom Sax-Parser auch die Attribute der Tags bergeben - die
            hier gemerkt und ebenfalls in endElement() weiterverarbeitet werden.
        """

        if name == "mapinfo":
            #Jetzt heissts aufpassen!
            self.attention= 1

        if self.attention:
            #Die Attribute des aktuellen Tags merken, sie werden auch in endElement verarbeitet
            self.attrs= attrs.copy()
        else:
            self.inhalt= ""
    
    
    def endElement(self, name):
        """
            Diese Methode wird immer aufgerufen, wenn ein schlieender Tag gefunden wurde. Da
            man dann davon ausgehen kann, dass eine von characters() gefundene Informationseinheit
            abgeschlossen wurde, werden die gelesenen Daten hier weiterverarbeitet, d.h., je nach
            Name des schlieenden Tags (der vom Parser bergeben wird) entsprechenden Variablen
            zugeordnet. 
            
            Wenn der Tag <mapinfo> gefunden und also geschlossen wurde, wird attention auf 0 
            gesetzt - der Rest der Datei sollte jetzt am Parser vorbeirauschen. ber die setvalues-
            Methode des GeoMap-Objekts (siehe da) werden die gelesenen Daten an das Objekt ber-
            geben. Der Rckgabewert dieser Methode wird in der Instanzvariable success gemerkt und
            kann dann von Maploader abgefragt werden.
        """
        
        if self.attention:
            self.inhalt= normalize_str(self.inhalt)

            if name== "map_name":
                self.mapname= self.inhalt
                self.id= getAttr(self.attrs, "id")
                self.pov= getAttr(self.attrs, "pov")
            elif name== "description":
                self.description= self.inhalt
            elif name== "maxwest":
                self.maxwest= self.inhalt
            elif name== "maxeast":
                self.maxeast= self.inhalt
            elif name== "maxsouth":
                self.maxsouth= self.inhalt
            elif name== "maxnorth":
                self.maxnorth= self.inhalt

            self.inhalt= ''

            if name== "mapinfo":
                #Das wars...
                self.attention= 0

                #Die gewonnenen Daten dem Kartenobjekt bergeben
                self.mapobject.setvalues(self.mapname, self.id, self.pov, self.maxwest, self.maxeast, self.maxnorth, self.maxsouth, self.description)
                self.success = 1

                    
                
class GeoMap_XMLObjectHandler(handler.ContentHandler):
    """ Diese Klasse wird vom Sax-XML-Parser benutzt. Sie soll die geometrischen Objekte aus
        einer Objektdatenbank im XML-Format fischen. Alle Methoden auer __init__() werden vom
        Sax-Parser bei bestimmten Ereignissen aufgerufen - siehe auch die Python-Dokumentation.
        
        Verwendung/ Aufruf der Methoden:
        Zur Erstellung eines neuen Handlers: __init__()
        Zum Verarbeiten von Inhalt (Zeichen innerhalb eines XML-Tags): characters()
        Beim Erreichen eines ffnenden Tags: startElement()
        Beim Erreichen eines schlieenden Tags: endElement()
        
        Verwendete Bibliotheken:
        Die Klasse muss von xml.sax.handler abgeleitet werden - also wird diese importiert.
        Um ein geometrisches Objekt erstellen zu knnen, mssen GeoObject und Point importiert
        werden. Die grammatischen Informationen eines Objekts werden in einem Nounlike-Objekt
        abgelegt.

        Die dieser Klasse erwartete XML-Struktur ist in odb.dtd beschrieben. Siehe dort und
        die dafr vorgesehene Dokumentation.                
    """

    def __init__(self, mapobject, name, id):
        """ Beim Erstellen eines Handlerobjekts muss ein Verweis auf das Kartenobjekt ber-
            geben werden, fr das die Objekte gelesen werden - nur so knnen sie nmlich auch
            gleich dort gespeichert werden. Aber diesen Verweis hat ja schon das Maploader-
            Objekt bekommen, und er wird nach hier durchgereicht (siehe die dortige Init-Methode).
            
            Auerdem knnen Name der zu ladenden Objekte oder ID des zu ladenden Objekts angegeben
            werden. Wird kein Parameter angegeben, werden alle Objekte geladen.
            
            Es wird eine Instanzvariable success initialisiert, an der Maploader feststellen kann,
            ob die Ausfhrung erfolgreich war. (Anfangswert = 0 = Misserfolg)
        """
                
        self.mapobject= mapobject

        #Erfolg berwachen
        self.success= 0

        #Ein paar Initialisierungen...
        self.inhalt= ''
        self.searchname= name
        self.searchid= id

        #Die Variblen fr die Objektdaten
        self.initDataVars()
        
        #Und ein paar Flags, die sagen, wo wir gerade sind
        self.attention= 0
        self.point_attention = 0
        self.dir_attention = 0

    
    def initDataVars(self):
        """ Initialisiert die Variablen, in denen die Daten eines Objekts gespeichert werden.
            Sehr praktisch, da es dann zum Aufrumen - falls mehrere Objekte geladen werden -
            gleich wieder aufgerufen werden kann.
        """

        self.objname= ""
        self.semantics= {}
        self.id= ""
        self.contained= []
        self.points= []
        self.direction= []
        self.kasus= ""
        self.numerus= ""
        self.genus= ""

        
    def characters(self, content):
        """ Wird immer dann aufgerufen, wenn Text zwischen Tags gefunden wurde - dieser wird
            in content bergeben. Die Methode macht nichts weiter, als sich den Text in einer
            Instanzvariable inhalt zu merken - verarbeitet wird der Inhalt immer beim Erreichen
            eines schlieenden Tags - also in endElement().
        """        
        
        self.inhalt+= content

    
    def startElement(self, name, attrs):
        """ Diese Methode wird immer aufgerufen, wenn ein ffnender Tag gefunden wurde. Sie 
            fungiert als eine Art Wchter. Ein neues Objekt beginnt immer mit einem <object>-
            Tag - dann wird attention auf 1 gesetzt. Und nur wenn attention= 1 ist, wird
            in endElement(), wo die Daten verarbeitet werden, berhaupt irgendetwas getan.
        
            Zwei weitere Flags regeln die richtige Verarbeitung der gefundenen Punkte: einmal
            als Punkte des Objekts, anschlieend als Punkte des Richtungsvektors. Da der
            Tag <points> zweimal vorkommt, sind tatschlich zwei Flags erforderlich. 
            DEBUG! Formatnderung?
           
            Dieser Methode werden vom Sax-Parser auch die Attribute der Tags bergeben - die
            hier gemerkt und ebenfalls in endElement() weiterverarbeitet werden.
        """
    
        #Ab diesem Tag wird es interessant, ein neues Objekt beginnt
        if name == "object":
            self.attention= 1

        if self.attention:
            #Flags zum richtigen Verarbeiten der gefundenen Punkte setzen
            if name== "points":
                if (not self.dir_attention): self.point_attention= 1
            if name== "direction":
                self.dir_attention= 1
                self.point_attention= 0
                              
            #Die Attribute merken, sie werden auch in endElement verarbeitet
            self.attrs= attrs.copy()
        else:
            self.inhalt= ""
    
    
    def endElement(self, name):
        """ Diese Methode wird immer aufgerufen, wenn ein schlieender Tag gefunden wurde. Da
            man dann davon ausgehen kann, dass eine von characters() gefundene Informationseinheit
            abgeschlossen wurde, werden die gelesenen Daten hier weiterverarbeitet, d.h., je nach
            Name des schlieenden Tags (der vom Parser bergeben wird) entsprechenden Variablen
            zugeordnet. 
            
            Wenn der Tag <object> gefunden und also geschlossen wurde, wird attention auf 0 
            gesetzt und die brigen Flags zurckgesetzt - das Objekt ist abgeschlossen und kann 
            gespeichert werden. Es wird angelegt und an die Objektliste der Karte angehngt.
            success wird auf 1 gesetzt.
        """
        
        if self.attention:
            self.inhalt= normalize_str(self.inhalt)
            
            if name== "obj_name":
                #Wenn nach einen Namen oder einer ID gesucht wird, muss hier verglichen werden
                if self.searchname:
                    if not str_equals(self.searchname, self.inhalt):
                        self.attention= 0
                
                self.objname= self.inhalt
                self.id= getAttr(self.attrs, "id")
                                    
                if self.searchid:
                    if self.searchid != self.id:
                        self.attention= 0    
                
                #Eine Liste der beinhalteten Objekte anlegen
                if self.attention and self.attrs.has_key("contained_objects"):
                    con= getAttr(self.attrs, "contained_objects")
                    self.contained= split(con)

            elif name== "semantics":
                #Semantikeintrge der Form "key=wert" in einem Dictionary speichern
                sem= split(self.inhalt)
                for x in sem:
                    sep= find(x, '=')
                    self.semantics[x[0:sep]]= x[sep+1:]
                                        
            elif name== "point":
                xy= split(self.inhalt)
                point= Point(0, float(xy[0]), float(xy[1]))
                
                #Je nach Flag zuordnen
                if self.point_attention:
                    self.points.append(point)
                elif self.dir_attention:
                    self.direction.append(point)

            elif name== "kasus":
                self.kasus= self.inhalt
            elif name== "numerus":
                self.numerus= self.inhalt
            elif name== "genus":
                self.genus= self.inhalt
                
            self.inhalt= ''

            #Ende eines Objekts? Dann abspeichern
            if name== "object":
                #Das wars...
                self.attention= 0
                self.point_attention= 0
                self.dir_attention= 0
    
                #Die grammtischen Informationen in ein Nounlike-Objekt stecken
                graminfo= Nounlike(self.objname, self.objname, self.kasus, self.numerus, self.genus)
                
                #Ein neues Geoobject anlegen
                gobj= GeoObject(self.objname, self.id, self.points, self.direction, self.semantics, self.contained, graminfo)

                #An die Liste der Kartenobjekte anhngen
                self.mapobject.addobject(gobj)
                self.success= 1

                #Noch ein bisschen aufrumen fr das nchste Objekt
                self.initDataVars()
        
        #Wenn self.attention= 0
        else:
            self.inhalt= ""
            
