# ****************************************************
# Name des Moduls: PrepLoader
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 09.01.2003
# ****************************************************


""" Die Preploader-Bibliothek wird von der Klasse PrepContainer verwendet, um Prpositionen
    und Modifikatoren aus externen Quellen laden zu knnen.
    
    Das Verstndnis dieser Bibliothek ist nicht ntig, um mit Preposition- oder PrepContainer
    -Objekten umgehen zu knnen. Es handelt sich um eine INTERNE Bibliothek.
    
    Die Bibliothek enthlt die Klasse Preploader. Um Daten laden zu knnen, wird von PrepContainer
    ein Preploader-Objekt erstellt. Diesem werden der Name des zu verwendenden Formats und 
    der Name der Quelle (Dateiname, URL) bergeben. Beim Aufruf der Methoden getpreps() oder
    getmods() wird dann je nach Format ein entsprechender Parser aufgerufen.
    
    Im Augenblick lediglich das XML-Format untersttzt. Es wird der von Python mitgelieferte
    Sax-Parser verwendet. Dem Parser muss ein Objekt bergeben werden, dessen Methoden zur
    Verarbeitung der gelesenen Daten aufgerufen werden. Die entsprechenden Klassen heien
    Prep_XMLHandler fr Prpositionen und Mod_XMLHandler fr Modifikatoren.
    
    Achtung: Wahrscheinlich erfordert der Sax-Parser eine Python-Version >= 2.2 !!!
"""


from string import *
from preposition import *
from berror import ParameterError
from btools import str_equals


# -------- XML -------------

#Fr den XML-Parser muss man Handler definieren, die die Daten verarbeitet -
#siehe Prep_XMLHandler
from xml.sax import handler

#Steuert den SAX-XML-Parser an
from xmlparser import XMLParser, normalize_str, getAttr


class Preploader:
    """ Mit Hilfe dieser Klasse knnen Prpositionen und Modifikatoren aus externen Quellen
        geladen werden
    """

    def __init__(self, cobject, sourcename, sourcetyp):
        """ Zum Erstellen eines Objekts werden folgende Parameter bentigt:
        
        ein Verweis auf den Prpositions-Container, fr den Daten geladen werden sollen (also
        eine Instanz von PrepContainer): cobject
        der Name der Quelle, aus der Daten geladen werden sollen: sourcename
        der Typ der Quelle, aus der geladen werden soll: sourcetyp
        
        Diese Daten werden gespeichert, ansonsten geschieht hier nichts mehr. """
        
        self.cobject= cobject
        self.sourcename= sourcename
        self.sourcetyp= sourcetyp.upper()


    def getpreps(self, name, id, projection):
        """ Ldt Prpositionen. Es knnen ein Name und/ oder eine ID angegeben werden -
            dann werden nur die Objekte mit diesem Namen oder das Objekt mit dieser ID geladen.
            Wird zustzlich ein Wert fr pov ("Grundriss" oder "Aufriss") angegeben, werden nur
            die Prpositionen geladen, die fr diese POV definiert sind.
            
            Je nach sourcetyp wird ein entsprechender Parser initialisiert und aufgerufen. War das
            Laden des Objekts/ der Objekte erfolgreich (diese werden direkt in den Prepcontainer 
            geschrieben), gibt die Methode 1 fr Erfolg zurck, 0 bei Misserfolg.
        """

        success = 0
        if str_equals(self.sourcetyp, "XML"):
            #Der Sax-Handler fuer gefundene Daten
            hd = Prep_XMLHandler(self.cobject, name, id, projection)

            #Einen Parser mit unserem Handler erschaffen...
            parser = XMLParser(hd, self.sourcename)

            #... und ihn schuften lassen
        
            parse_success = self.xmlparse(parser)
            
            if parse_success:
                #Noch nachsehen, ob der Handler Probleme hatte...
                success = hd.success

        return success


    def getmods(self, name, id, projection):
        """ Ldt Modifikatoren. Es knnen ein Name und/ oder eine ID angegeben werden -
            dann werden nur die Objekte mit diesem Namen oder das Objekt mit dieser ID geladen.
            Wird zustzlich ein Wert fr pov ("Grundriss" oder "Aufriss") angegeben, werden nur
            die Modifikatoren geladen, die fr diese POV definiert sind.
            
            Je nach sourcetyp wird ein entsprechender Parser initialisiert und aufgerufen. War das
            Laden des Objekts/ der Objekte erfolgreich (diese werden direkt in den Prepcontainer 
            geschrieben), gibt die Methode 1 fr Erfolg zurck, 0 bei Misserfolg.
        """

        success= 0
        if str_equals(self.sourcetyp, "XML"):
            #Der Sax-Handler fuer gefundene Daten
            hd= Mod_XMLHandler(self.cobject, name, id, projection)

            #Einen Parser mit unserem Handler erschaffen...
            parser= XMLParser(hd, self.sourcename)

            #... und ihn schuften lassen
            parse_success= self.xmlparse(parser)
            
            if parse_success:
                #Noch nachsehen, ob der Handler Probleme hatte...
                success= hd.success

        return success


    def xmlparse(self, parser):
        """ Ruft den XML-Parser auf. Der Aufruf wird hier zentralisiert, um die Fehlerbehandlung
            nicht mehrmals schreiben zu mssen - der Parser muss nmlich sowohl in getpreps() als
            auch in getmods() aufgerufen werden. Dort wird auch ein Parserobjekt mit dem richtigen
            Handler initialisiert und hier in der Variable parser bergeben.
            
            Gibt 1 bei Erfolg, 0 bei Misserfolg zurck. """

        success = 1
        try:
            parser.parse()
        except ParameterError:
            #Die selbstdefinierten Fehler wollen wir zu sehen bekommen
            raise
        except:
            success = 0
            
        return success            


class Prep_XMLHandler(handler.ContentHandler):
    """ Diese Klasse wird vom Sax-XML-Parser benutzt. Sie soll Prpositionen aus
        einer Datenbank im XML-Format fischen. Alle Methoden auer __init__() werden vom
        Sax-Parser bei bestimmten Ereignissen aufgerufen - siehe auch die Python-Dokumentation.
        
        Verwendung/ Aufruf der Methoden:
        Zur Erstellung eines neuen Handlers: __init__()
        Zum Verarbeiten von Inhalt (Zeichen innerhalb eines XML-Tags): characters()
        Beim Erreichen eines ffnenden Tags: startElement()
        Beim Erreichen eines schlieenden Tags: endElement()
        
        Verwendete Bibliotheken:
        Die Klasse muss von xml.sax.handler abgeleitet werden - also wird diese importiert.
        Um eine Prposition erstellen zu knnen, muss Preposition importiert werden.

        Die in dieser Klasse erwartete XML-Struktur richtet sich nach der in praeps.dtd
        vorgeschriebenen Struktur. Siehe auch dort und die dafr vorhandene Dokumentation.
    """

    def __init__(self, cobject, name, id, projection):
        """ Beim Erstellen eines Handlerobjekts muss ein Verweis auf den PrepContainer ber-
            geben werden, fr den die Prps gelesen werden - nur so knnen sie nmlich auch
            gleich dort gespeichert werden. Aber diesen Verweis hat ja schon das Preploader-
            Objekt bekommen, und er wird nach hier durchgereicht (siehe die dortige Init-Methode).
            
            Auerdem knnen Name und/ oder ID zu ladenden Prps angegeben werden. Mit
            projection kann man die Auswahl einschrnken, es werden nur die Prps geladen, auf die Name
            oder id UND projection zutrifft. Wird kein Parameter angegeben, werden alle Prpositionen
            geladen.
            
            Es wird eine Instanzvariable success initialisiert, an der Preploader feststellen kann,
            ob die Ausfhrung erfolgreich war. (Anfangswert = 0 = Misserfolg)
        """
                
        self.cobject= cobject

        #Erfolg berwachen
        self.success= 0

        #Ein paar Initialisierungen...
        self.inhalt= ''
        self.searchname= name
        self.searchid= id
        self.searchprojection= projection
        self.aktprojection= ""
        self.attention= 0
        
        #... und Variablen fr die Daten einer Prposition
        self.initPrepVars()
        self.initBufferVars()


    def initPrepVars(self):
        """" Initialisiert die Variablen, in denen die gelesenen Prpositionsdaten abgelegt werden """
        
        #Allgemeine Informationen
        self.name= ""
        self.id= ""
        self.bemerkung= ""
        self.povs= []

        #Eine Liste fr die Bufferinformationsblcke
        self.prepbuffer= []

        #Grammatische Informationen        
        self.discont= []
        self.syntax= ""
        self.wortart= ""
        self.kasus= ""
        self.kombination= []


    def initBufferVars(self):
        """ Initialisiert die Variablen, in denen die Daten zur Bufferberechnung gespeichert werden.
            Fr jede Prposition kann es mehrere Stze dieser Informationen geben. Sie werden in der
            Liste self.prepbuffer gespeichert (siehe intPrepVars() und endElement()
        """

        #allgemeine Informationen zur Bufferberechnung        
        self.flaechentyp= ""
        self.kontakt= ""
        self.dyn_stat= ""
        self.schraffur= ""
       
        #Typ-A-spezifische Informationen
        self.bb_seite= 0
        self.relevant_obj_depth= 0
        self.inner_distance= 0
        self.outer_distance= 0
        self.angle= 0
        self.angle_expand= 0        
        
        
    def characters(self, content):
        """ Wird immer dann aufgerufen, wenn Text zwischen Tags gefunden wurde - dieser wird
            in content bergeben. Die Methode macht nichts weiter, als sich den Text in einer
            Instanzvariable inhalt zu merken - verarbeitet wird der Inhalt immer beim Erreichen
            eines schlieenden Tags - also in endElement().
        """        

        self.inhalt+= content
        
    
    def startElement(self, name, attrs):
        """ Wird aufgerufen, wenn ein ffender Tag gefunden wurde """
    
        #berprfen, ob wir in einem interessanten Teil des Dokuments sind
        if name == "preposition":
            self.attention = 1

            #Wenn wir whlerisch sind, was den POV angeht, wird hier nachgesehen,
            #ob unsere Aufmerksamkeit tatschlich gerechtfertigt ist
            if self.searchprojection:
                if not str_equals(self.searchprojection, self.aktprojection):
                    self.attention= 0

        #Welche Sichtweise haben wir denn gerade?
        if name == "horizontal_projection":
            self.aktprojection= "grundriss"
        if name == "upright_projection":
            self.aktprojection= "aufriss"
            
                
        if self.attention:
            #Die Attribute des aktuellen Tags merken, sie werden auch in endElement verarbeitet
            self.attrs= attrs.copy()
        else:
            self.inhalt= ""
    
    
    def endElement(self, name):
        """
            Diese Methode wird immer aufgerufen, wenn ein schlieender Tag gefunden wurde. Da
            man dann davon ausgehen kann, dass eine von characters() gefundene Informationseinheit
            abgeschlossen wurde, werden die gelesenen Daten hier weiterverarbeitet, d.h., je nach
            Name des schlieenden Tags (der vom Parser bergeben wird) entsprechenden Variablen
            zugeordnet. 
            
            Wenn der Tag <praeposition> gefunden und also geschlossen wurde, wird attention auf 0 
            gesetzt und die brigen Flags zurckgesetzt - das Objekt ist abgeschlossen und kann 
            gespeichert werden. Es wird angelegt und an die Prpliste des Containers angehngt.
            success wird auf 1 gesetzt.
        """
        
        if self.attention:
            self.inhalt= normalize_str(self.inhalt)
            if name== "name":
                self.name= self.inhalt
                self.id= getAttr(self.attrs, "id")
                self.projection= self.aktprojection

                #Wenn nach einen Namen oder einer ID gesucht wird, muss hier verglichen werden
                if self.searchname:
                    if not str_equals(self.searchname, self.name):
                        self.attention= 0
                        
                if self.searchid:
                    if self.searchid != self.id:
                        self.attention= 0    
                                   
            #Grammatische Infos
            elif name== "token":
                #Die diskontinuierlichen Elemente der Praep in einer Liste speichern
                self.discont.append(self.inhalt)

            elif name== "syntax":
                self.syntax= self.inhalt
            elif name== "part_of_speech":
                self.wortart= self.inhalt
            elif name== "case":
                self.kasus= self.inhalt
            elif name== "combinations":
                z = self.inhalt.replace(' ', '')
                self.kombination = z.split(',')
                
            #Infos zur erzeugten Flche
            elif name== "area":
                self.flaechentyp += getAttr(self.attrs, "type")
            elif name== "contact":
                self.kontakt = getAttr(self.attrs, "type")
            elif name== "point_of_view":
                self.povs.append(getAttr(self.attrs, "type"))
            elif name== "dyn_stat":
                self.dyn_stat = getAttr(self.attrs, "type")
            elif name== "coloring":
                self.schraffur = getAttr(self.attrs, "intensity")
    
            #Typ-A-Info
            elif name== "bb_side":
                self.bb_seite= getAttr(self.attrs, "side")
            elif name== "relevant_object_depth":
                self.relevant_obj_depth = int(self.inhalt)
            elif name== "inner_distance":
                self.inner_distance = int(self.inhalt)
            elif name== "outer_distance":
                self.outer_distance = int(self.inhalt)
            elif name== "angle":
                self.angle = int(self.inhalt)
                self.angle_expand= getAttr(self.attrs, "expand")
                if str_equals(self.angle_expand, "yes"):
                    self.angle_expand= 1
                else:
                    self.angle_expand= 0
                        
            #...und die obligatorische Bemerkung ;)
            elif name== "notes":
                self.bemerkung= self.inhalt


            self.inhalt= ""

            #Ende einer Prep oder eines Prepbuffers? Dann abspeichern
            if name== "prep_buffer":                
                pb= PrepBuffer()
                pb.setGeneralBufferValues(self.flaechentyp, self.kontakt, self.dyn_stat, self.schraffur)
                if str_equals(self.flaechentyp, "A"):
                    pb.setTypABufferValues(self.bb_seite, self.relevant_obj_depth, self.inner_distance, self.outer_distance, self.angle, self.angle_expand)
                else:
                    #(Noch) unbekannter Prpositionstyp
                    raise ParameterError("Prep_XMLHandler.endElement", 6)

                self.prepbuffer.append(pb)
                self.initBufferVars()

            if name== "preposition":
                #Das wars...
                self.attention= 0
                
                #Ein neues Prpositionsobjekt anlegen
                nprep= Preposition(self.name, self.id, self.projection, self.povs, self.bemerkung)
                nprep.setGrammarValues(self.discont, self.syntax, self.wortart, self.kasus, self.kombination)
                nprep.prepbuffer= self.prepbuffer                

                #An die Liste der Kartenobjekte anhngen
                self.cobject.cpreps.append(nprep)
                self.success= 1

                #Noch ein bisschen aufrumen fr das nchste Objekt
                self.initPrepVars()

        #Wenn self.attention= 0
        else:
            self.inhalt= ""
            


class Mod_XMLHandler(handler.ContentHandler):
    """ Diese Klasse wird vom Sax-XML-Parser benutzt. Sie soll Modifikatoren aus
        einer Datenbank im XML-Format fischen. Alle Methoden auer __init__() werden vom
        Sax-Parser bei bestimmten Ereignissen aufgerufen - siehe auch die Python-Dokumentation
        und die Doku der Klasse Prep_XMLHandler.

        Da Modifikatoren bei der Bufferberechnung noch nicht bedacht sind, werden hier frs
        erste nur die grammatischen Infos geladen, die der Tagger bentigt.

        Es wird die Klasse Modifier in preposition bentigt, um ein Modifikatorobjekt anlegen
        zu knnen.

        Die in dieser Klasse erwartete XML-Struktur richtet sich nach der in praeps.dtd
        vorgeschriebenen Struktur. Siehe auch dort und die dafr vorhandene Dokumentation.
    """

    def __init__(self, cobject, name, id, projection):
        """ Beim Erstellen eines Handlerobjekts muss ein Verweis auf den PrepContainer ber-
            geben werden, fr den die Mods gelesen werden - nur so knnen sie nmlich auch
            gleich dort gespeichert werden. Aber diesen Verweis hat ja schon das Preploader-
            Objekt bekommen, und er wird nach hier durchgereicht (siehe die dortige Init-Methode).
            
            Auerdem knnen Name und/ oder ID der zu ladenden Mods angegeben werden. Mit
            projection kann man die Auswahl einschrnken, es werden nur die Mods geladen, auf die name
            oder id UND projection zutrifft. Wird kein Parameter angegeben, werden alle Modifikatoren
            geladen.
            
            Es wird eine Instanzvariable success initialisiert, an der Preploader feststellen kann,
            ob die Ausfhrung erfolgreich war. (Anfangswert = 0 = Misserfolg)
        """
                
        self.cobject= cobject

        #Erfolg berwachen
        self.success= 0

        #Ein paar Initialisierungen...
        self.inhalt= ''
        self.searchname= name
        self.searchid= id
        self.searchprojection= projection
        self.aktprojection= ""
        self.attention= 0
        
        #... und Variablen fr die Daten eines Modifikators
        self.initDataVars()
        

    def initDataVars(self):
        """" Initialisiert die Variablen, in denen die gelesenen Modifikatordaten abgelegt werden """
        
        #Allgemeine Informationen
        self.name= ""
        self.id= ""

        #Grammatische Infos
        self.pos= ""

        
    def characters(self, content):
        """ Wird immer dann aufgerufen, wenn Text zwischen Tags gefunden wurde - dieser wird
            in content bergeben. Die Methode macht nichts weiter, als sich den Text in einer
            Instanzvariable inhalt zu merken - verarbeitet wird der Inhalt immer beim Erreichen
            eines schlieenden Tags - also in endElement().
        """        
        
        self.inhalt+= content
        
    
    def startElement(self, name, attrs):
        """ Wird aufgerufen, wenn ein ffender Tag gefunden wurde """
    
        #berprfen, ob wir in einem interessanten Teil des Dokuments sind
        if name == "modifier":
            self.attention = 1
            if self.searchprojection:
                if not str_equals(self.searchprojection, self.aktprojection):
                    self.attention= 0

        #Welche Sichtweise haben wir denn gerade?
        if name == "horizontal_projection":
            self.aktprojection= "grundriss"
        if name == "upright_projection":
            self.aktprojection= "aufriss"
                
        if self.attention:
            #Die Attribute merken, sie werden auch in endElement verarbeitet
            self.attrs= attrs.copy()
        else:
            self.inhalt= ""
    
    
    def endElement(self, name):
        """ Diese Methode wird immer aufgerufen, wenn ein schlieender Tag gefunden wurde. Da
            man dann davon ausgehen kann, dass eine von characters() gefundene Informationseinheit
            abgeschlossen wurde, werden die gelesenen Daten hier weiterverarbeitet, d.h., je nach
            Name des schlieenden Tags (der vom Parser bergeben wird) entsprechenden Variablen
            zugeordnet. 
            
            Wenn der Tag </modifier> gefunden und also geschlossen wurde, wird attention auf 0 
            gesetzt und die brigen Flags zurckgesetzt - das Objekt ist abgeschlossen und kann 
            gespeichert werden. Es wird angelegt und an die Modliste des Containers angehngt.
            success wird auf 1 gesetzt.
        """
        
        if self.attention:
            self.inhalt= normalize_str(self.inhalt)
            if name== "name":
                self.name= self.inhalt
                self.id= getAttr(self.attrs, "id")
                
                #Wenn nach einen Namen oder einer ID gesucht wird, muss hier verglichen werden
                if self.searchname:
                    if not str_equals(self.searchname, self.name):
                        self.attention= 0
                        
                if self.searchid:
                    if self.searchid != self.id:
                        self.attention= 0    
                                   
            #Grammatische Infos
            elif name== "part_of_speech":
                self.pos= self.inhalt
            
            self.inhalt= ""

            #Ende eines Modifikators? Dann abspeichern...
            if name== "modifier":
                #Das wars...
                self.attention= 0
                
                #Einen neuen Modifikator anlegen
                nmod= Modifier(self.name, self.id, self.pos)
                
                #An die Liste der Kartenobjekte anhngen
                self.cobject.cmods.append(nmod)
                self.success= 1

                #Noch ein bisschen aufrumen fr das nchste Objekt
                self.initDataVars()

        #Wenn self.attention= 0
        else:
            self.inhalt= ""