# ****************************************************
# Name des Moduls: Preposition
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 10.01.2003
# ****************************************************


""" In diesem Modul befinden sich Klassen, um die Daten von Prpositionen und Modifikatoren
    zu speichern.

    In einem Objekt der Klasse Preposition knnen allgemeine und grammatische Informationen
    einer Prposition abgelegt werden (fr eine genauere Beschreibung dieser Informationen siehe
    auch die Dokumentation der Prpositionsdatenbank bzw. von praeps.dtd). Auerdem hat jedes
    Prpositionsobjekt eine Liste prepbuffer, in der Instanzen der Klasse Prepbuffer gespeichert
    werden knnen (siehe nchster Absatz).

    In einem Objekt der Klasse Prepbuffer knnen die Informationen gespeichert werden, die
    zur Berechnung eines Buffers fr eine Prposition bentigt werden. Da eine Prposition mehrere
    Buffer erzeugen kann (bei neben sind es z.B. zwei), knnen mehrere Instanzen dieser Klasse
    in der prepbuffer-Liste eines Preposition-Objekts abgelegt werden. Nhere Informationen zu
    den hier gespeicherten Parametern finden Sie in der Dokumentation von praeps.dtd bzw. auch
    bei der Beschreibung der Bufferalgorithmen.

    In einem Objekt der Klasse Modifier knnen die Daten eines Modifikators gespeichert werden.
    Diese sind im Augenblick nur rudimentr, da die Verwendung von Modifikatoren zwar vorge-
    sehen, aber noch kaum implementiert ist. Sie knnen im Augenblick in Prpositionalphrasen
    verwendet werden, werden bei der Berechnung des Buffers jedoch ignoriert.
"""

from btools import str_equals


class Preposition:
    
    def __init__(self, name, id, projection, povs, comment= ""):
        """ Legt ein neues Prpositionsobjekt an und initialisiert ein paar Variablen,
            die allgemeingltige und grammatikalische Informationen fr alle Prpositions-
            typen speichern. Die Werte der grammatischen Parameter knnen mit der Methode
            setGrammarValues() gesetzt werden.
            
            Auerdem wird eine Liste fr PrepBuffer angelegt, in denen die Informationen
            zur Bufferberechnung gespeichert werden knnen.
        """
                       
        #Allgemeine Informationen
        self.name= name
        self.id= id
        self.bemerkung= comment
        self.projection= projection
        self.povs= povs

        #Grammatische Informationen        
        self.discont= []
        self.syntax= ""
        self.wortart= ""
        self.kasus= ""
        self.kombination= []

        #Eine Liste mit PrepBuffern
        self.prepbuffer= []

    
    def setGrammarValues(self, discont, syntax, wortart, kasus, kombination):
        """ Speichert fr die verwendete Prposition die bergebenen grammatischen Informationen. """
        
        self.discont= discont
        self.syntax= syntax
        self.wortart= wortart
        self.kasus= kasus
        self.kombination= kombination


    def __repr__(self):
        """ bergibt die Prposition als String, der dann ausgegeben werden kann."""
        
        ausgabe= "Praeposition: "+ self.name
        ausgabe+= "\nBemerkung: "+ self.bemerkung
        ausgabe+= "\nID: "+ self.id
                
        ausgabe+= "\n\nGrammatische Informationen:"
        if len(self.discont) > 1:
            ausgabe+= "\nDie einzelnen Elemente der diskontinuierlichen Praeposition sind:"
            for x in self.discont:
                ausgabe+= "\n   " + x
        else:
            ausgabe+= "\nDie Prposition ist nicht diskontinuierlich."

        ausgabe+= "\nWortart: "+ self.wortart
        ausgabe+= "\nKasus: "+ self.kasus
        ausgabe+= "\nSyntax der Prpositionalphrase: "+ self.syntax
        ausgabe+= "\nKombinationsmglichkeiten mit Modifikatoren:"
        for x in self.kombination:
                ausgabe+= "\n   " + x
                
        ausgabe+= "\n\nInformationen zur Bufferberechnung:\n"
        ausgabe+= "\nGeometrische Sichtweise: "+ self.projection
        ausgabe+= "\nMgliche Betrachterstandpunkte:"
        for x in self.povs:
                ausgabe+= "\n   " + x
        ausgabe+= "\n\nFolgende(r) Buffer werden (wird) erzeugt:\n\n"
        for x in self.prepbuffer:
            ausgabe+= x.__repr__()
            ausgabe += "\n\n"
                    
        return ausgabe


class PrepBuffer:
    """ In dieser Klasse knnen die Informationen abgelegt werden, die zur Berechnung
        eines Buffers (eine Flche, die den Gltigkeitsbereich einer Prposition aus-
        drckt) notwendig sind. Es wird unterschieden zwischen allgemeinen Informationen,
        die (wahrscheinlich) fr alle Buffertypen ntig ist (zur Beschreibung der Buffertypen
        siehe die allgemeine Konzeptbeschreibung des Projekts), und buffertypspezifischen
        Informationen. Diese knnen mit den verschiedenen set...()-Funktion gesetzt und
        verndert werden. Eine nhere Beschreibung der einzelnen Parameter findden Sie
        in der Beschreibung der praeps.dtd oder in der Beschreibung der Bufferalgorithmen.
    """
        

    def __init__(self):
        """ Initialisiert ein neues PrepBuffer-Objekt mit den fr alle Buffertypen
            gltigen Variablen. Zum Setzen dieser Informationen kann die Methode
            setGeneralBufferValues() verwendet werden.
        """
                       
        #(Vermutlich) allgemeine Informationen zur Bufferberechnung
        self.flaechentyp= ""
        self.kontakt= ""
                
        #Informationen zur Schraffur des Buffers
        self.dyn_stat= ""
        self.schraffur= ""    
        
        
    def setGeneralBufferValues(self, flaechentyp, kontakt, dyn_stat, schraffur):
        """ Speichert fr die verwendete Prposition die Informationen zur Bufferberechnung,
            die fr alle Typen von Bufferflchen gleich sind. 
        """
        
        #Informationen zur Bufferberechnung        
        self.flaechentyp= flaechentyp
        self.kontakt= kontakt
        
        #Informationen zur Schraffur des Buffers
        self.dyn_stat= dyn_stat
        self.schraffur= schraffur
        

    def setTypABufferValues(self, bb_seite= "", relevant_obj_depth= 0, i_distance= 0, o_distance= 0, angle= 0, angle_expand= 0):
        """ Legt die Typ-A-spezifischen Variablen zur Bufferberechnung fest """

        self.bb_seite= int(bb_seite)
        self.bb_seite -= 1
        self.relevant_obj_depth= int(relevant_obj_depth)
        self.inner_distance= int(i_distance)
        self.outer_distance= int(o_distance)
        self.angle= int(angle)
        self.angle_expand= int(angle_expand)


    def __repr__(self):
        """ bergibt die Informationen als String, der dann ausgegeben werden kann. """

        ausgabe= ""                
        if str_equals(self.flaechentyp, "A"):
            ausgabe+= "Es entsteht eine Flche vom Typ A."
                            
            ausgabe+= "\nDie Flaeche entsteht auf der BoundingBox-Seite "+ str(self.bb_seite)
            ausgabe+= "\nRelevante Objekttiefe: "+ str(self.relevant_obj_depth)
            ausgabe+= "\nEntfernung der inneren und aeusseren Linie vom Objekt: " + str(self.inner_distance) + " / " + str(self.outer_distance)
            ausgabe+= "\nWinkel (expand): "+ str(self.angle) + " (" + str(self.angle_expand)+")"

        ausgabe+= "\nKontakt: "+ self.kontakt
        ausgabe+= "\nBufferschraffur ist "+self.dyn_stat+" und "+self.schraffur+"."
                    
        return ausgabe


class Modifier:
    """ Mit Hilfe dieser Klasse kann ein sog. Modifikator gespeichert werden - das
        ist ein Wort, das den Geltungsbereich einer Prposition modifiziert: LINKS
        neben der Kirche. Die Berechnung von Modifikatoren ist noch nicht implementiert,
        diese Klasse wurde aber schon geschrieben, damit der Eingabestring-Tagger mit
        Modifikatoren umgehen kann. Daher werden im Augenblick nur rudimentre Infos
        gespeichert.
    """

    def __init__(self, name, id, pos):

        self.name = name
        self.id = id
        self.pos = pos

    def __repr__(self):

        ausgabe = "\nModifikator " + self.name + " mit ID " + self.id
        ausgabe += "\nPOS: " + self.pos + "\n"
        return ausgabe
