# ********************************************************
# Name des Moduls: Ptag
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
# unter der Verwendung eines Prototyps von Jan-Ole Esleben
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 12.01.2003
# ********************************************************


""" Dieses Modul enthlt einen Parser fr Prpositionalphrasen (Ptag). Er
	hat die Aufgabe, fr eine bergebene Prpositionalphrase die IDs der
	entsprechenden Eintrge in der verwendeten Prpositionsdatenbank
	(fr Prps und Modifikatoren) und in der jeweiligen Karte (fr Objekte)
	zurckzugeben. Dabei mssen gelegentlich Ambiguitten aufgelst werden:
	so kann z.B. eine Prposition in unterschiedlichem Kasus eine jeweils
	andere Bedeutung haben (in), oder eine Wort kann sowohl als Prposition
	als auch als Modifikator verwendet werden (links, rechts), oder...

	Auer der Klasse Ptag befindet sich noch die Klasse Lookup im Modul.
	Mit ihrer Hilfe schlgt der Parser in den verschiedenen Datenbanken nach.

	Verwendete Module:
	geomap, prepcontainer und gram: werden von Lookup zum Nachschlagen in den
		verschiedenen DBs bentigt
	btools: fr div. Stringvergleiche
"""

from geomap import GeoMap
from prepcontainer import PrepContainer
from gram import *
from btools import str_equals, strlist_equals, listcontains
from berror import DataError, ParameterError, InternalError, ParserError
import copy, re



class Lookup:
	""" Der Parser (Ptag) erstellt sich am Anfang ein Objekt dieser Klasse,
		um mit seiner Hilfe  die Elemente des Eingabestrings in den verschiedenen
		Datenbanken nachzuschlagen und zu identifizieren
	"""

	def __init__(self, prepdb, preptyp, objdb, objtyp, gramdb, gramtyp):
		""" Es werden die Namen und Typen aller drei Datenbanken bergeben. Die
			Prpositionen und Blends werden gleich vollstndig geladen; die Blends
			werden sowieso alle bentigt, und auf die Prpositionen wird so oft
			zugegriffen, dass ein wiederholtes Parsen der DB das Nachschlagen ver-
			langsamt. Fr die Artikel aus der Grammatik-DB und die Karte werden+
			vorerst lediglich Loader-Objekte initialisiert.
		"""
		
		#Diverse Daten laden bzw. Loader generieren. Grammatik...
		self.gramloader = Gramloader(gramdb, gramtyp)

		#...Blends...
		self.blends = []
		if self.gramloader.getblends():
			self.blends = self.gramloader.blends
		else:
			raise DataError("Lookup.init", 1, gramdb, "Blends")
		
		#...Objekte...
		self.map = GeoMap(objdb, objtyp)
		if self.map.getparams():
			self.projection = self.map.pov
		else:
			raise DataError("Lookup.init", 2, objdb, "Kartenparameter")

		#... und  Prpositionen :-)
		self.prepcon = PrepContainer(prepdb, preptyp)

		#Alle Prpositionen laden (wg. der Geschwindigkeit)
		if not self.prepcon.getpreps(projection= self.projection):
			raise DataError("Lookup.init", 3, prepdb, "")	

	
	def replace_blends(self, s):
		""" Ersetzt im String s das Vorkommen aller geladenen Blends mit ihren
			Ersetzungen.
		"""
		for i in self.blends:
			s= i.replace(s)	
		
		return s

	def ask_prepdb(self, el):
		""" berprft, ob der String el eine Prposition aus der aktuellen Praep-DB
			ist. Gibt prep_string() ("Prep") zurck, wenn ja, None, wenn nein.
		"""

		erg= None
		if self.prepcon.searchpreps(name= el):
			erg= prep_string()

		return erg			

	def ask_mod(self, el):
		""" berprft, ob der String el ein Modifikator aus der aktuellen Prp-DB
			ist. Gibt mod_string() ("Mod") zurck, wenn ja, None, wenn nein.
		"""

		erg= None
		if self.prepcon.getmods(name= el):
			erg= mod_string()

		return erg

	def ask_gramdb(self, el):
		""" berprft, ob der String el ein Artikel aus der aktuellen Gram-DB
			ist. Gibt art_string() ("Art") zurck, wenn ja, None, wenn nein.
		"""
		
		erg = None
		if self.gramloader.getdeterminer(el):
			erg= art_string()

		return erg			

	def ask_objdb(self, el):
		""" berprft, ob der String el ein Objekt aus der aktuellen Karte
			ist. Gibt noun_string() ("Noun") zurck, wenn ja, None, wenn nein.
		"""
		
		erg= None
		if self.map.getobjects(name= el):
			erg= noun_string()

		return erg

	def expand_dets(self, el):
		""" Gibt eine Liste mit Tupeln zurck, die alle grammatikalischen Beschreibungen
			von el (ein Artikel) enthalten: [(Kasus, Numerus, Genus, ID), ...] """

		artikel = []
		if self.gramloader.getdeterminer(el):
			artikel = self.gramloader.artikel
		else:
			raise DataError("ptag.py: Lookup.expand_dets", "5", self.gramloader.sourcename, "")
		
		forms= []
		for i in artikel:
			s= i.kasus, i.numerus, i.genus, "*"
			forms.append(s)
			
		return forms			

	def expand_preps(self, el):
		""" Gibt eine Liste mit Tupeln zurck, die alle grammatikalischen Beschreibungen
			von el (einem Prpositionswort) beinhalten: [(Kasus, Numerus, Genus, ID), ...] """

		erg = self.prepcon.searchpreps(name= el)
		forms= []
		for i in erg:
			s= i.kasus, "*", "*", i.id
			forms.append(s)
			
		return forms

	def expand_prep_id(self, prep_id):
		""" Gibt ein Tupeln fr die Prep mit der angegebenen ID zurck, das die
			grammatikalischen Beschreibung der Prep enthlt: [(Kasus, Numerus, Genus, ID), ...] """

		erg = self.prepcon.getpreps(id= prep_id)
		s= erg[0].kasus, "*", "*", self.prepcon.cpreps[0].id
		return s

	def expand_nouns(self, el):
		""" Gibt eine Liste mit Tupeln zurck, die alle grammatikalischen Beschreibungen
			von el (einem Nomen) zurckgeben: [(Kasus, Numerus, Genus, ID), ...] """

		self.map.cobjects= []
		if self.map.getobjects(name= el):
			forms= []
			for i in self.map.cobjects:
				#Kasus wird rausgelassen, da immer Nominativ...
				s= "*", i.graminfos.numerus, i.graminfos.genus, i.id
				forms.append(s)
		else:
			raise DataError("ptag.py: Lookup.expand_nouns", "4", self.map.loader.sourcename, "Objekt "+ el)
			
		return forms				

	def expand_mod(self, el):
		""" Gibt einen leeren Unifikationsstring fr den Modifikator zurck, fuegt aber
			noch die ID hinten an """

		s = ""
		self.prepcon.cmods= []
		if self.prepcon.getmods(name= el):
			s= ("*","*","*", self.prepcon.cmods[0].id)
		else:
			raise DataError("ptag.py: Lookup.expand_mods", "6", self.prepcon.loader.sourcename, "Modifikator "+ el)		
		return s

	def get_syntax(self, prep_id):
		""" Gibt den Syntaxstring und die gltigen Mods der bergebenen Prposition zurck """

		erg = self.prepcon.searchpreps(id= prep_id)
		return erg[0].syntax, erg[0].kombination

	def get_all_discontinous_preps(self):
		""" Gibt eine Liste aller diskontinuierlichen Prpositionen in der Prp-DB zurck """
		
		dis_preps = []
		for p in self.prepcon.cpreps:
			if len(p.discont) > 1:
				dis_preps.append(p)

		return dis_preps			


def mod_string():
	return "Mod"

def is_mod(el):
	if mod_string() == el: return 1
	return None

def prep_string():
	return "Prep"

def is_prep(el):
	if prep_string() == el: return 1
	return None

def noun_string():
	return "Noun"

def is_noun(el):
	if noun_string() == el: return 1
	return None

def art_string():
	return "Art"

def is_art(el):
	if art_string() == el: return 1
	return None


class PTag:
	""" Dieser Parser fr Prpositionalphrasen soll die Elemente des Eingabestrings
		identifizieren und ihre IDs in den verschiedenen Datenbanken zurckgeben. Wie
		der Name schon vermuten lt, handelt es sich mehr um einen Tagger als einen
		Parser. Die einzelnen Wrter des Eingabestrings werden in den verschiedenen
		Datenbanken nachgeschlagen und getagt. Wenn alle Wrter identifiziert werden
		konnten, wird versucht, ihre grammatikalischen Attribute zu unifizieren. Ge-
		lingt auch das, wird die Wortreihenfolge noch mit der Vorgabe verglichen, die
		die erkannte Prposition fr eine gltige Prpositionalphrase fordert. Ist
		auch dieser Vergleich erfolgreich, wird die Eingabe akzeptiert.

		Untersttzte Phnomene:
		- diskontinuierliche Prpositionen (z. B. an..entlang)
		- disambiguiert gleiche Prpositionen in unterschiedlichen Kasus: in die Kirche/
		  in der Kirche
		- unterscheidet zwischen gleichlautenden Prpositionen und Modifikatoren
		- Verschmelzungsphnomene wie "an dem --> am" werden bercksichtigt

		Nicht untersttzte Phnomene:
		- Nomendeklination wird nicht bercksichtigt. Objektnamen mssen immer in
		  exakt der Form eingegeben werden, in der sie in den Karten stehen, also
		  "links des Rathaus" statt "links des Rathauses"
		- Spezialfall "im Zentrum": das "im" wrde von der Blends-Ersetzung zu "in dem"
		  gemacht werden, "in dem Zentrum" steht aber wahrscheinlich nicht in der
		  Prp-DB...
	"""
	
	def __init__(self, s, prepdb, preptyp, objdb, objtyp, gramdb, gramtyp):
		""" In s wird der Eingabestring bergeben. Ansonsten mssen Namen und Typ der
			drei Datenbanken angegeben werden, die der Parser verwenden soll.

			War der Parsevorgang erfolgreich, befinden sich die Ergebnisse anschliessend
			in der Liste self.results. (Normalerweise sollte es nur ein Ergebnis geben,
			aber theoretisch sind ambige Prpositionalphrasen denkbar.) Ein einzelnes
			Parse-Ergebnis ist wiederum eine Liste - mit folgendem Aufbau:

				1.	einer Liste der IDs aller Modifikatoren
				2.	die ID der Prposition
				3.	eine Liste der IDs aller Objekte

			Beispiel: [[3], 12, [5]]. Da mehrere Modifikatoren und auch mehrere Objekte
			pro Prpositionalphrase mglich sind ("links oben in der Ecke", "zwischen A
			und B"), befinden sich diese Elemente in weiteren Sublisten.

			Wenn kein Parse gefunden werden konnte, wird ein Fehler erzeugt. Das ist
			letztendlich Geschmackssache - man knnte auch anhand der leeren results-Liste
			feststellen, dass es kein Ergebnis gab, und den Fehler dann an bergeordneter
			Stelle erzeugen.
		"""

		self.results= []
		if s:
		
			#Ein Lookup-Objekt erschaffen ;)
			self.lookup= Lookup(prepdb, preptyp, objdb, objtyp, gramdb, gramtyp)
									
			#Den Eingabestring vorbereiten
			s = self.lookup.replace_blends(s)
			self.phrase = re.split(r"[^\w]+", s)

			#Mgliche Interpretationen der Eingabe finden
			self.struct = []
			self.unknown_elements = 0
			self.get_structures()

			#Wieviele als Prep getagte Elemente haben wir denn?
			self.preps = self.all_preps()
			self.anzahl_preps= len(self.preps)

			if self.unknown_elements == 0:
				self.identify_singleprep()
			if self.unknown_elements > 0 or self.anzahl_preps > 1:
				self.identify_discontpreps()

			#Wenns keine Ergebnisse gibt, einen Fehler generieren. Ist Geschmacks-
			#sache, wers nicht mag, solls streichen
			if self.results == []:
				raise ParserError("Ptag.init", 3)

				
	def get_structures(self):
		""" Versucht, die einzelen Wrter des Eingabestrings anhand der verschiedenen
			DBs zu identifizieren. Wenn nicht alle Elemente identifiziert werden konnten,
			heit dass noch nicht, dass die Eingabe ungltig war - es knnte sich auch
			noch um eine diskontinuierliche Prposition handeln.
		"""
		
		for el in self.phrase:
			matches = []

			#Ist das Wort eine Prep und/ oder ein Modifikator?
			pos = self.lookup.ask_prepdb(el)
			if pos:
				matches.append(pos)

			pos= self.lookup.ask_mod(el)
			if pos:
				matches.append(pos)

			#Ansonsten knnte es noch ein Artikel ODER ein Objekt sein
			if not matches:
				pos = self.lookup.ask_gramdb(el)
				if not pos:
					pos= self.lookup.ask_objdb(el)
				if pos:
					matches.append(pos)

			if not matches:
				self.unknown_elements += 1
			else:
				self.struct.append(matches)


	def identify_singleprep(self):
		""" Versucht eine Interpretation der Eingabe unter der Annahme, das nur
			ein einzelnes Prpositionswort darin vorkommt - also keine diskontinuierliche
			Prep vorliegt. Wenn es mehrere als Prposition getagte Wrter gibt, wird
			versucht, sie jeweils mit einer Ausnahme (eben der tatschlichen Prp) als
			Modifikator zu verstehen.

			Fr alle gefundenen Interpretationen wird dann versucht, die Elemente mit-
			einander zu unifizieren und mit dem Syntaxstring der Prposition zu vergleichen.
			Alle Interpretationen, die diese Tests berstehen, werden als gltige Ergebnisse
			gewertet.
		"""
		interpretations = []

		if self.anzahl_preps == 1:
			#Der einfachste Fall. Eine eindeutige Interpretation:
			inter = []
			for i in self.struct:
				inter.append(i[0])
			interpretations.append(inter)

		elif self.anzahl_preps > 1:
			#Feststellen, ob sich alle (bis vielleicht auf eine) Preps auch
			#als Mods verstehen lassen
			mods= self.all_mods()
			if len(mods) == (self.anzahl_preps -1):
				#Eine Prep kann nicht Mod sein - auch eine eindeutige Interpretation
				inter = []
				for i in self.struct:
					if listcontains(i, mod_string()):
						inter.append(mod_string())
					else:
						inter.append(i[0])
				interpretations.append(inter)

			elif len(mods) == self.anzahl_preps:
				#Alle Preps lassen sich auch als Mods lesen. Also gibt es genau soviele
				#mglichen Interpretationen
				for pos in self.preps:
					inter = []
					for i in range(len(self.struct)):
						if pos == i:
							inter.append(prep_string())
						else:
							if listcontains(self.struct[i], mod_string()):
								inter.append(mod_string())
							else:
								inter.append(self.struct[i][0])
					interpretations.append(inter)

		#Fr alle gefundenen Interpretationen berprfen, ob auf Basis dieser
		#Interpretation die grammatischen Kategorien der Wrter unifizieren
		for inter in interpretations:
			uni = self.unify(inter)
			
			#Fr alle gltigen Unifikationen noch berprfen, ob der
			#Syntaxstring der ermittelten Prep mit dem erkannten berstimmt
			for j in uni:
				ids = self.get_ids(j, inter)
				if self.check_syntax(ids[1], inter):
					self.results.append(ids)


	def identify_discontpreps(self):
		""" berprft, ob sich in der Eingabe eine diskontinuierliche Prposition
			verbirgt und berprft, falls das der Fall ist, die Syntax der Eingabe.
			Alle Interpretationen, deren grammatikalischen Attribute unifizieren
			und deren Wortreihenfolge mit der Vorgabe der erkannten Prp ber-
			einstimmt, werden als gltige Ergebnisse gemerkt.
		"""

		discont_preps = self.lookup.get_all_discontinous_preps()

		for dp in discont_preps:
			ppos = copy.deepcopy(self.preps)
			matches = 0

			#berprfen, ob alle Wrter der diskontin. Prep in der Eingabe vorkommen
			for el in dp.discont:
				for i in range(self.anzahl_preps):
					if self.phrase[self.preps[i]] == el:
						matches += 1
						ppos.remove(self.preps[i])
						break;

			#Wenn alle Teile der diskontinuierlichen Prep gefunden wurden...
			if matches == len(dp.discont):

				#berprfen, ob alle weiteren Preps als Mods gelesen werden knnen
				mods_ok = 1
				for i in ppos:
					if not listcontains(self.struct[i], mod_string()):
						mods_ok = 0
						break

				#Falls alles in Ordnung ist, Syntaxstring fr den Vergleich basteln
				if mods_ok:
					inter = []
					for i in range(len(self.struct)):
						if listcontains(self.struct[i], prep_string()):
							if i in ppos:
								inter.append(mod_string())
							else:
								inter.append(prep_string())
						else:
							inter.append(self.struct[i][0])

					#Versuchen, die Wrter nach dieser Interpretation zu unifizieren
					uni = self.unify(inter, dp.id)
					
					#Fr alle gltigen Unifikationen noch berprfen, ob der
					#Syntaxstring der ausgewhlten Prep mit dem erkannten berstimmt
					for j in uni:
						ids = self.get_ids(j, inter)
						if self.check_syntax(ids[1], inter):
							self.results.append(ids)					
										

	def all_preps(self):
		""" Ermittelt die Listenpositionen(!) aller als Prposition getagten Elemente
			im Tagstring (self.struct).
		"""
		r = []
		for i in xrange(len(self.struct)):
			if listcontains(self.struct[i], prep_string()): r.append(i)
		return r


	def all_mods(self):
		""" Ermittelt die Listenpositionen(!) aller als Modifikator getagten Elemente
			im Tagstring (self.struct).
		"""
		r = []
		for i in xrange(len(self.struct)):
			if listcontains(self.struct[i], mod_string()): r.append(i)
		return r


	def unify(self, inter, prep_id = None):
		""" Erstellt anhand einer eindeutigen Liste von Tags (einer mglichen
			Interpretation der Eingabe, in inter) eine Liste von Tupeln aller
			mglichen grammatikalischen Interpretationen der so getagten Wrter.
			Anschliessend wird versucht, die Sublisten miteinander zu unifizieren.
			Wenn es sich um bei der Interpretation um eine diskontinuierliche
			Prposition handelt, muss zustzlich noch die Prep-ID angegeben werden,
			um die richtige Prposition in der DB nachschlagen zu knnen.
		"""
			
		corlist = [[]]
		prep_ok = 0
		for i in xrange(len(inter)):
			#Von den verschiedenen Wrtern werden die gram. Attribute nachgeschlagen und
			#in Unifikationsstrings gesteckt, die hier zu Tupeln zusammengefasst werden
			
			if is_art(inter[i]) or is_prep(inter[i]) or	is_noun(inter[i]):

				if is_art(inter[i]):
					forms= self.lookup.expand_dets(self.phrase[i])
				elif is_prep(inter[i]):
					if not prep_id:
						forms= self.lookup.expand_preps(self.phrase[i])
					else:
						#Das ist der Zweig fr diskont. Preps...
						if prep_ok:
							forms= [('*','*','*','*')]
						else:
							forms = [self.lookup.expand_prep_id(prep_id)]
							prep_ok = 1
							
				elif is_noun(inter[i]):
					forms= self.lookup.expand_nouns(self.phrase[i])
								
				cortemp = corlist[:]
				corlist = []
				for form in forms:
					for el in cortemp:
						el1 = el[:]
						el1.append(form)
						corlist.append(el1)

			elif is_mod(inter[i]):
				#Wenns aber ein Modifikator ist, der mit allem unifiziert...
				form = self.lookup.expand_mod(self.phrase[i])
				for el in corlist:
					el.append(form)
			else:
				#Hoppla - ein unbekanntes Wort im Eingabestring
				raise ParserError("Ptag.unify", 2)

		#Jetzt sollen die Tupel durch Unifizieren reduziert werden...
		corlist = self.delete_invalids(corlist)
		return corlist


	def delete_invalids(self, corlist):
		""" Streicht aus der Liste mit den grammatischen Expansionen der Eingabewoerter all
			die Sublisten heraus, die nicht unifizieren """

		nlist= []
		for uni in corlist:
			#uni ist eine Liste von Tupeln, die miteinander unifizieren sollen
			if len(uni) < 2:
				raise InternalError("Ptag.delete_invalids", 2)
	
			#Anzahl der Elemente, die ein Tupel enthlt (-1 wg. der ID)
			n_elements= len(uni[0]) - 1

			fehler = 0			
			for el in range(n_elements):
				#Die Werte aller Tupel an Pos el sammeln und vergleichen
				values = []
				for t in uni:
					if t[el] != "*":
						values.append(t[el])

				n_values = len(values) - 1
				for i in range(n_values):
					if not str_equals(values[i],values[i+1]):
						fehler = 1
						break;

				if fehler == 1:
					break;
			
			#Falls beim Unifizieren kein Fehler aufgetreten ist, Element bernehmen
			if not fehler:
				nlist.append(uni)
	
		return nlist


	def get_ids(self, uni, inter):
		""" Extrahiert die IDs aus dem unifizierten Ergebnis. Erzeugt eine Liste,
			bestehend aus den folgenden Elementen:
				1.	einer Liste der IDs aller Modifikatoren
				2.	die ID der Prposition
				3.	eine Liste der IDs aller Objekte
		"""

		mod_ids= []
		prep_id= ""
		obj_ids= []
		
		id_pos= len(uni[0])-1
		for i in range(len(uni)):
			
			if is_mod(inter[i]):
				mod_ids.append(uni[i][id_pos])
			elif is_prep(inter[i]):
				id = uni[i][id_pos]
				if id != '*':
					prep_id = uni[i][id_pos]
			elif is_noun(inter[i]):
				obj_ids.append(uni[i][id_pos])

		return [mod_ids, prep_id, obj_ids]				
			

	def check_syntax(self, id, inter):
		""" berprft, ob die erkannte syntaktische Struktur (in inter) mit der Vorgabe
			der erkannten Prposition (die mit ID id) bereinstimmt. berprft weiterhin,
			ob die (evtl.) verwendeten Modifikatoren fr diese Prep erlaubt	sind. Gibt
			wahr zurck, wenn alles in Ordnung ist, sonst falsch.
		"""
		success = 0

		#Syntaxstring fr die erkannte Prep laden
		s, legal_mods = self.lookup.get_syntax(id)
		syntax = s.split()
		syntax_elements = len(syntax)
		inter_elements = len(inter)
		
		#Der geladene Syntaxstring muss mindestens so lang sein wie der erkannte...
		if syntax_elements >= inter_elements:
			
			#Der geladene Syntaxstring kann optionale Elemente enthalten und somit
			#lnger sein als der erkannte Syntaxstring. Also zwei Zhler...
			syntax_pos = 0
			inter_pos = 0

			all_matches = 0
			oblig_matches = 0

			#Anzahl der Elemente bestimmen, die im Syntaxstring obligatorisch sind
			min_matches = syntax_elements
			for i in syntax:
				#Optionale Elemente beginnen mit einem Stern
				if i[0] == "*":
					min_matches -= 1
			
			while syntax_pos < syntax_elements and inter_pos < inter_elements:
				optional = 0
											
				#Ist das Element optional?
				if syntax[syntax_pos][0] == "*":
					cmp = syntax[syntax_pos][1:]
					optional = 1
				else:
					cmp = syntax[syntax_pos]

				#Das Element in den beiden Strings vergleichen
				if cmp == inter[inter_pos]:
					all_matches += 1
					if not optional:
						oblig_matches += 1
				else:
					#Wenn die Elemente nicht gleich waren, aber das gesuchte Element optional
					#gewesen ist, drfen wir in der erkannten Struktur noch nicht vorrcken
					if optional:
						inter_pos -= 1

				syntax_pos += 1
				inter_pos += 1

			#Der Vergleich war erfolgreich, wenn:
			#		alle obligatorischen Elemente des Syntaxstrings gematched wurden
			#		alle Wrter der Eingabe gematched wurden
			if ((oblig_matches == min_matches) and (all_matches == inter_elements)):
				success = 1

		#Jetzt noch schauen, ob nur legale Mods verwendet wurden
		if success and (legal_mods != ["*"]):
			mod_pos = []
			for i in xrange(len(inter)):
				if inter[i] == mod_string(): mod_pos.append(i)
		
			for i in mod_pos:
				if not listcontains(legal_mods, self.phrase[i]):
					success = 0
					break;
				
		return success
		
